from langchain.document_loaders import PyMuPDFLoader
from langchain.retrievers import ArxivRetriever


def scrape_pdf_with_pymupdf(url) -> str:
    """Scrape a pdf with pymupdf

    Args:
        url (str): The url of the pdf to scrape

    Returns:
        str: The text scraped from the pdf
    """
    loader = PyMuPDFLoader(url)
    doc = loader.load()
    return str(doc)


def scrape_pdf_with_arxiv(query) -> str:
    """Scrape a pdf with arxiv
    default document length of 70000 about ~15 pages or None for no limit

    Args:
        query (str): The query to search for

    Returns:
        str: The text scraped from the pdf
    """
    retriever = ArxivRetriever(load_max_docs=2, doc_content_chars_max=None)
    docs = retriever.get_relevant_documents(query=query)
    return docs[0].page_content



import logging
import uuid
from datetime import datetime
from typing import Dict, List, Optional

from fastapi import HTTPException
from sqlmodel import delete, desc, select
from sqlmodel.ext.asyncio.session import AsyncSession

from mtmai.db.db import get_async_session
from mtmai.models.chat import ChatElement, ChatFeedback, ChatStep, ChatThread

logger = logging.getLogger()


async def get_user_threads(
    session: AsyncSession,
    user_id: str,
    thread_id: str = None,
    limit: int = 100,
    skip: int = 0,
) -> Optional[list[dict]]:
    """
    获取用户的 chat threads , 包含 steps 和 feedbacks
    """
    query = (
        select(ChatThread)
        .where((ChatThread.userId == user_id) | (ChatThread.id == thread_id))
        .order_by(desc(ChatThread.created_at))
        .limit(limit)
        .offset(skip)
    )

    result = await session.exec(query)
    user_threads = result.scalars().all()

    if not user_threads:
        return None

    thread_ids = [thread.id for thread in user_threads]

    steps_feedbacks_query = (
        select(ChatStep, ChatFeedback)
        .join(ChatFeedback, ChatStep.id == ChatFeedback.for_id, isouter=True)
        .where(ChatStep.thread_id.in_(thread_ids))
        .order_by(ChatStep.created_at)
    )

    steps_feedbacks_result = await session.exec(steps_feedbacks_query)
    steps_feedbacks = steps_feedbacks_result.all()

    threads_dict = []

    # convert to chainlit  user_threads format
    for thread in user_threads:
        thread_dict = {
            "id": str(thread.id),
            "createdAt": thread.created_at.isoformat(),
            "name": thread.name,
            "userId": str(thread.userId),
            "userIdentifier": thread.userIdentifier,
            "tags": thread.tags,
            "metadata": thread.meta,
            "steps": [],
        }

        for step, feedback in steps_feedbacks:
            if str(step.thread_id) == str(thread.id):
                step_dict = {
                    "id": str(step.id),
                    "name": step.name,
                    "type": step.type,
                    "threadId": str(step.thread_id) if step.thread_id else None,
                    "parentId": str(step.parent_id) if step.parent_id else None,
                    "streaming": step.streaming,
                    "waitForAnswer": step.wait_for_answer,
                    "isError": step.is_error,
                    "metadata": step.meta,
                    "tags": step.tags,
                    "input": step.input,
                    "output": step.output,
                    "createdAt": step.created_at.isoformat(),
                    "start": step.start.isoformat() if step.start else None,
                    "end": step.end.isoformat() if step.end else None,
                    "generation": step.generation,
                    "showInput": step.show_input,
                    "language": step.language,
                    "indent": step.indent,
                    "feedback": {
                        "value": feedback.value if feedback else None,
                        "comment": feedback.comment if feedback else None,
                    },
                }
                thread_dict["steps"].append(step_dict)

        threads_dict.append(thread_dict)

    return threads_dict


async def get_steps_by_thread(thread_id: uuid.UUID | str):
    if isinstance(thread_id, str):
        thread_id = uuid.UUID(thread_id)
    async with get_async_session() as session:
        query = select(ChatStep).where(ChatStep.thread_id == thread_id)
        result = await session.exec(query)
        return result.all()


async def create_chat_element(
    chat_element_dict: dict,
    user_id: str | None = None,
):
    async with get_async_session() as session:
        new_item = ChatElement.model_validate(chat_element_dict)
        if not new_item.mime:
            new_item.mime = "application/octet-stream"
        session.add(new_item)
        await session.commit()
        return new_item


async def get_chat_element(
    thread_id: uuid.UUID | str,
    element_id: uuid.UUID | str,
):
    if isinstance(thread_id, str):
        thread_id = uuid.UUID(thread_id)
    if isinstance(element_id, str):
        element_id = uuid.UUID(element_id)

    async with get_async_session() as session:
        query = select(ChatElement).where(
            ChatElement.threadId == thread_id, ChatElement.id == element_id
        )
        result = await session.exec(query)
        return result.first()


async def is_thread_author(username: str, thread_id: str):
    thread_author = await get_thread_author(thread_id)

    if not thread_author:
        raise HTTPException(status_code=404, detail="Thread not found")

    if thread_author != str(username):
        raise HTTPException(status_code=401, detail="Unauthorized")
    else:
        return True


async def create_thread(
    *,
    thread_id: uuid.UUID | None = None,
    userId: str,
    meta_data: dict = None,
    tags: list[str] = [],
    name: str = None,
):
    _name = name or f"chat-{datetime.now().strftime('%m%d-%H-%M')}"

    async with get_async_session() as session:
        new_thread = ChatThread(
            id=thread_id,
            userIdentifier=str(userId),
            meta=meta_data or {},
            tags=tags or [],
            userId=userId,
            name=_name,
        )
        session.add(new_thread)
        await session.commit()
        await session.refresh(new_thread)
        return new_thread


async def update_thread(
    *,
    thread_id: uuid.UUID | str,
    name: Optional[str] = None,
    userId: uuid.UUID | str = None,
    metadata: Optional[Dict] = None,
    tags: Optional[List[str]] = None,
):
    if isinstance(thread_id, str):
        thread_id = uuid.UUID(thread_id)
    if isinstance(userId, str):
        userId = uuid.UUID(userId)

    async with get_async_session() as session:
        existing_thread = await get_thread(thread_id)
        if existing_thread is None:
            existing_thread = await create_thread(
                thread_id=thread_id,
                userId=userId,
                meta_data=metadata,
                tags=tags or [],
                name=name,
            )
        if name is not None:
            existing_thread.name = name
        if userId is not None:
            existing_thread.userIdentifier = str(userId)
            existing_thread.userId = userId
        if metadata is not None:
            existing_thread.meta = metadata
        if tags is not None:
            existing_thread.tags = tags
        await session.commit()


async def get_thread(thread_id: uuid.UUID | str):
    if isinstance(thread_id, str):
        thread_id = uuid.UUID(thread_id)

    async with get_async_session() as session:
        query = select(ChatThread).where(ChatThread.id == thread_id)
        result = await session.exec(query)
        return result.one_or_none()


async def get_thread_author(thread_id: uuid.UUID | str):
    result = await get_thread(thread_id)
    if result:
        return result.userIdentifier
    else:
        return None


async def delete_thread(thread_id: uuid.UUID | str):
    # Delete feedbacks/elements/steps/thread
    if isinstance(thread_id, str):
        thread_id = uuid.UUID(thread_id)

    async with get_async_session() as session:
        await session.exec(
            delete(ChatFeedback).where(ChatFeedback.thread_id == thread_id)
        )
        await session.exec(
            delete(ChatElement).where(ChatElement.thread_id == thread_id)
        )
        await session.exec(delete(ChatStep).where(ChatStep.thread_id == thread_id))
        await session.exec(delete(ChatThread).where(ChatThread.id == thread_id))
        await session.commit()



from prefect import flow


@flow(log_prints=True)
def hello_world(name: str = "world", goodbye: bool = False):
    print(f"Hello {name} from Prefect! 🤗")

    if goodbye:
        print(f"Goodbye {name}!")


if __name__ == "__main__":
    # creates a deployment and stays running to monitor for work instructions
    # generated on the server

    hello_world()



import base64
import getpass
import json
import logging
import os
import random
import shlex
import subprocess

# import pwd
import tempfile
import threading
import time
import traceback
from pathlib import Path

# from stem.util import term
from urllib.parse import urlparse

import requests
from mtlibs import mtutils
from mtlibs.mtutils import ranstr
from mtlibs.process_helper import exec, is_tool
from requests.packages.urllib3.exceptions import InsecureRequestWarning

logger = logging.getLogger()

# 全局设置，忽略ssl证书校验警告
requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

# 外网IP检查间隔（秒）
TOR_CHECK_IP_INTERVAL = 10


def ensureIntor():
    """确保网络在tor里面"""
    tor_socks_port = os.environ.get("TOR_SOCKS_PORT", "0.0.0.0:9050")
    controll_port = os.environ.get("TOR_CONTROLL_PORT", "127.0.0.1:9051")
    controll_password = os.environ.get("TORCONTROLL_PASSWORD", "my_password")
    b64_hs_ed25519_secret_key = os.environ.get("ONIONKEY", None)
    if mtutils.is_tcp_listen(controll_port):
        logger.info("####tor controll %s 运行中" % controll_port)
        return True
    else:
        set_iptables()
        tor_proc = TorProc()
        print("开始启动tor进程", flush=True)
        is_tor_connected = tor_proc.start(
            socks_port=tor_socks_port,
            hiddenservice_ports=["80 80"],
            controll_port=controll_port,
            controll_password=controll_password,
            b64_hs_ed25519_secret_key=b64_hs_ed25519_secret_key,
            wait=True,
        )
        print("tor进程启动完毕", flush=True)
        if is_tor_connected:
            print(term.format("TOR 已经连接上", term.Color.GREEN), flush=True)
            print(
                term.format(
                    "TOR hidden dir: %s" % tor_proc.hidden_dir, term.Color.GREEN
                ),
                flush=True,
            )

            if tor_proc.onion_name:
                os.environ["MAIN_ONION"] = tor_proc.onion_name
            print(
                term.format("onion:  %s" % tor_proc.onion_name, term.Color.GREEN),
                flush=True,
            )
            print(
                term.format(
                    "private key:  %s" % tor_proc.private_key, term.Color.GREEN
                ),
                flush=True,
            )
            threading.Thread(target=ip_protect, daemon=True).start()
            return True
        else:
            return False


def ip_protect():
    """
    后台线程，不断检测IP是否处于安全状态
    """
    current_user = getpass.getuser()
    # curl --insecure https://116.202.120.181/api/ip
    urllist = ["https://116.202.120.181/api/ip", "https://check.torproject.org/api/ip"]
    count = 1
    total_seconds = 0
    while True:
        url = urllist[count % len(urllist)]
        # logger.info("准备检测：{}".format(url))
        try:
            t1 = time.time()
            response = requests.get(url, verify=False)
            resultObj = json.loads(response.content.decode())
            if resultObj["IsTor"] == "False":
                logger.error("错误，外网IP泄露")
                exit(-128)
            else:
                current_seconds = time.time() - t1
                total_seconds += current_seconds
                ava_seconds = total_seconds / count
                logger.info(
                    "u:%s %.2f/%.2f ip:%s %s"
                    % (current_user, current_seconds, ava_seconds, resultObj["IP"], url)
                )
        except requests.exceptions.ConnectionError as e:
            print(e)
            logger.info(f"ip_protect -> user:{current_user}, 连接失败 {url}")
        time.sleep(TOR_CHECK_IP_INTERVAL)
        count += 1


class TorProc:
    """表示一个可控的Tor进程"""

    def __init__(self):
        self.connected = False
        self._hidden_dir = "hidden_service"

    def getHostname(self):
        """获取隐藏服务的域名"""
        with open(self.hidden_dir + "/hostname") as f:
            return f.read().strip()

    def getSocksPort(self):
        """获取当前tor进程的socks5代理端口"""
        return self.socks_port

    def checkIpSafe(self):
        proxies = {
            "http": "socks5://{}:{}".format("127.0.0.1", self.socks_port),
            "https": "socks5://{}:{}".format("127.0.0.1", self.socks_port),
        }
        res = requests.get("https://check.torproject.org/api/ip", proxies=proxies)
        return res.content.decode().index('"IsTor":true') > 0

    async def stop(self):
        """停止tor进程"""
        # TODO: 停止tor进程

    def start(
        self,
        socks_port=9050,
        b64_hs_ed25519_secret_key=None,
        hiddenservice_ports=["80 80"],
        controll_password="my_password",
        controll_port="9051",
        user="debian-tor",
        wait=True,
    ):
        """
        socks_port :0 为随机端口
        返回值:
        [是否成功,实际实用的socks端口号]
        """
        # 获取外部代理
        HTTP_PROXY = os.environ.get("HTTP_PROXY") or os.environ.get("http_proxy")

        EXTERNAL_PROXY_HOST = None
        EXTERNAL_PROXY_PORT = None
        if HTTP_PROXY:
            logger.info(f"外部代理设置： {HTTP_PROXY}")
            http_proxy_uri = urlparse(HTTP_PROXY)
            EXTERNAL_PROXY_HOST = http_proxy_uri.hostname
            EXTERNAL_PROXY_PORT = http_proxy_uri.port

        if socks_port == 0:
            socks_port = random.randrange(40001, 49999)
        self.socks_port = socks_port
        self.hiddenservice_ports = hiddenservice_ports
        toruser = pwd.getpwnam(user)
        # toruser = "debian-tor"
        if not is_tool("tor"):
            exec("sudo apt install -y tor")

        self.tmpdirname = tempfile.gettempdir() + "/tor_" + ranstr(20)
        self.hidden_dir = os.path.join(self.tmpdirname, self._hidden_dir)
        Path(self.tmpdirname).mkdir(mode=0o700, exist_ok=True)
        # logger.info(f"更改目录owner {self.hidden_dir}, {toruser.pw_uid}, {toruser.pw_gid}")
        # for root, dirs, files in os.walk(self.tmpdirname):
        #     for momo in dirs:
        #         # os.chown(momo, 502, 20)
        #         os.chown(momo, toruser.pw_uid, toruser.pw_gid)
        os.chown(self.tmpdirname, toruser.pw_uid, toruser.pw_gid)

        torcc_lines = []
        torcc_lines.append(f"SocksPort 0.0.0.0:{socks_port}")
        torcc_lines.append(f'DataDirectory "{self.tmpdirname}"')
        torcc_lines.append(f"ControlPort {controll_port}")
        torcc_lines.append("CookieAuthentication 1")
        torcc_lines.append("VirtualAddrNetworkIPv4 10.192.0.0/10")
        torcc_lines.append("AutomapHostsOnResolve 1")
        torcc_lines.append("TransPort 0.0.0.0:9040")
        torcc_lines.append("DNSPort 0.0.0.0:5353")
        if EXTERNAL_PROXY_HOST and EXTERNAL_PROXY_PORT:
            torcc_lines.append(
                f"Socks5Proxy {EXTERNAL_PROXY_HOST}:{EXTERNAL_PROXY_PORT}"
            )
        # tor 使用外部代理
        # torargs.append('HTTPProxy 192.168.239.1:7890')
        # torargs.append('HTTPProxyAuthenticator username:password')
        # torargs.append('HTTPProxyAuthenticator username:password')
        # torargs.append('Socks5Proxy host:port')
        # torargs.append('Socks5ProxyUsername username')
        # torargs.append('Socks5ProxyPassword password')
        # torargs.append('ExitNodes ru')

        if self.hiddenservice_ports:
            # tor 会使用现成的hs_ed25519_secret_key文件生成公钥和域名. 这样就相当于预先指定了onion域名
            # 否则,会重新生成新的私钥和新的域名
            Path(self.hidden_dir).mkdir(mode=0o700, exist_ok=True)

            logger.info(
                f"hidden_dir: {self.hidden_dir}, {toruser.pw_uid}, {toruser.pw_gid}"
            )
            os.chown(self.hidden_dir, toruser.pw_uid, toruser.pw_gid)
            torcc_lines.append(f'HiddenServiceDir "{self.hidden_dir}"')

            for port in self.hiddenservice_ports:
                torcc_lines.append(f'HiddenServicePort "{port}"')

            # 必须配置了端口，才能配置私钥，要不然tor不能正常启动。
            if b64_hs_ed25519_secret_key:
                filepath = self.hidden_dir + "/hs_ed25519_secret_key"
                Path(filepath).touch(mode=0o700)
                with open(filepath, "wb") as f:
                    f.write(base64.b64decode(b64_hs_ed25519_secret_key))
                os.chown(filepath, toruser.pw_uid, toruser.pw_gid)

        # 明文参数
        # torargs.append("--hash-password 'my_password'")
        # 密码使用hash
        # 这个hash 对应的密码："my_password"
        torcc_lines.append(
            "HashedControlPassword 16:E600ADC1B52C80BB6022A0E999A7734571A451EB6AE50FED489B72E3DF"
        )

        torcc_path = "/tmp/torcc_tor_" + str(random.randrange(0, 999999999999))
        logger.info(f"torcc_path: {torcc_path}")
        torcc_content = "\n".join(torcc_lines)
        with open(torcc_path, "w") as fd:
            fd.write(torcc_content)

        # os.chown(torcc_path, toruser)
        os.chown(torcc_path, toruser.pw_uid, toruser.pw_gid)

        logger.info(f"torcc_path: {torcc_path} \n torcc_content: {torcc_content}\n")
        cmdline = f"sudo -u debian-tor tor -f {torcc_path}"
        logger.info(f"[exec]{cmdline}")

        self.process = subprocess.Popen(shlex.split(cmdline), stdout=subprocess.PIPE)

        connect_result = self.waite_tor_connect()
        if connect_result == False:
            logger.info("TOR 启动失败")
            return False
        else:
            self.onion_name = self.getHostname()
            self.private_key = self.getPrivateKey_b64()
            self.dataDirectory = self.tmpdirname
            return True

        # else:
        #     threading.Thread(target=self.waite_tor_connect).start()

    def getHostname(self):
        """获取隐藏服务的实际域名"""
        hostnameFile = self.hidden_dir + "/hostname"
        if Path(hostnameFile).exists():
            with open(hostnameFile) as f:
                return f.read().strip()
        else:
            return None

    def getPrivateKey_b64(self):
        """获取隐藏服务对应的私钥的base64编码的字符"""
        keyfile = self.hidden_dir + "/hs_ed25519_secret_key"
        if Path(keyfile).exists():
            with open(keyfile, "rb") as f:
                return base64.b64encode(f.read()).decode()
        else:
            return None

    def wait(self):
        """等待直到连接上"""
        while not self.connected:
            time.sleep(1)
        return

    def waite_tor_connect(self):
        """处理tor进程的输出，当成功连接，或者失败时返回"""
        try:
            while True:
                if self.process.stdout:
                    line = self.process.stdout.readline().decode().strip()
                    logger.debug(f"tor->{line}")
                    if line.strip().find("Bootstrapped 100%") > 0:
                        self.connected = True
                        return True
                    elif (
                        line.strip().find("Catching signal TERM, exiting cleanly.") > 0
                    ):
                        logger.info("[tor exited]")
                        return False
                    elif line.strip().find("[err]") > 0:
                        logger.error(f"tor 连接出错 {line.strip()}")
                        return False

                if self.process.stderr:
                    line = self.process.stderr.readline().decode()
                    logger.error(f"tor->[error]{line}")
                    return False

                if self.process.poll() is not None:
                    break

        except Exception as e:
            logger.error(traceback.format_exc(e))
            return False


def set_iptables():
    """
    设置IPtables
    1: 本地自动透明代理
    2：vpn自动透明代理（跟openvpn搭配使用）
    3: 特定用户进程能真实IP连接外网
    4: 只允许特定端口入站。
    """
    print("set_iptables called", flush=True)

    _tor_uid = pwd.getpwnam("debian-tor").pw_uid
    # 另外一个高权限的非tor用户
    _nottor_uid = "567"
    # Tor's TransPort
    _trans_port = "9040"
    # Tor's DNSPort
    _dns_port = "5353"
    # Tor's VirtualAddrNetworkIPv4
    _virt_addr = "10.192.0.0/10"
    # Your outgoing interface
    _out_if = "eth0"
    # LAN destinations that shouldn't be routed through Tor
    # _non_tor="127.0.0.0/8 10.0.0.0/8 172.16.0.0/12 192.168.0.0/16 10.12.12.0/24 172.17.0.0/16 10.18.0.0/16"
    _non_tor = "127.0.0.0/8 10.0.0.0/8 172.16.0.0/12 192.168.0.0/16 172.17.0.0/16"
    # Other IANA reserved blocks (These are not processed by tor and dropped by default)
    _resv_iana = "0.0.0.0/8 100.64.0.0/10 169.254.0.0/16 192.0.0.0/24 192.0.2.0/24 192.88.99.0/24 198.18.0.0/15 198.51.100.0/24 203.0.113.0/24 224.0.0.0/4 240.0.0.0/4 255.255.255.255/32"
    # 允许入站端口
    _input_ports = "80 8080 49090 41890 41080 48811"
    os.system("iptables -F && iptables -t nat -F")
    os.system(
        f"""iptables -t nat -A OUTPUT -d {_virt_addr} -p tcp -m tcp --tcp-flags FIN,SYN,RST,ACK SYN -j REDIRECT --to-ports {_trans_port}"""
    )

    # nat dns requests to Tor
    exec(
        f"iptables -t nat -A OUTPUT -d 127.0.0.1/32 -p udp -m udp --dport 53 -j REDIRECT --to-ports {_dns_port}"
    )

    # Don't nat the Tor process, the loopback, or the local network
    exec(f"iptables -t nat -A OUTPUT -m owner --uid-owner {_tor_uid} -j RETURN")
    exec(f"iptables -t nat -A OUTPUT -m owner --uid-owner {_nottor_uid} -j RETURN")
    exec("iptables -t nat -A OUTPUT -o lo -j RETURN")

    # 内网网段
    for _lan in _non_tor.split(" "):
        exec(f"iptables -t nat -A OUTPUT -d {_lan} -j RETURN")
    for _iana in _resv_iana.split(" "):
        exec(f"iptables -t nat -A OUTPUT -d {_iana} -j RETURN")
    # Redirect all other pre-routing and output to Tor's TransPort
    exec(
        f"iptables -t nat -A OUTPUT -p tcp -m tcp --tcp-flags FIN,SYN,RST,ACK SYN -j REDIRECT --to-ports {_trans_port}"
    )

    # 允许入站的端口
    for _input_port in _input_ports.split(" "):
        exec(
            f"iptables -A INPUT -i {_out_if} -p tcp --dport {_input_port} -m state --state NEW -j ACCEPT"
        )

    # 关闭其他入站端口
    exec("iptables -A INPUT -m state --state ESTABLISHED -j ACCEPT")
    exec("iptables -A INPUT -i lo -j ACCEPT")

    # Allow INPUT from lan hosts in $_non_tor
    # 允许内网入站连接
    for _lan in _non_tor.split(" "):
        exec(f"iptables -A INPUT -s {_lan} -j ACCEPT")

    # iptables -A INPUT -j DROP
    # 暂时接受所有入站请求，毕竟是在容器内，本身就有端口的限制。
    # exec("iptables -A INPUT -j DROP")
    exec("iptables -A INPUT -j ACCEPT")
    # *filter FORWARD
    exec("iptables -A FORWARD -j DROP")

    # *filter OUTPUT
    exec("iptables -A OUTPUT -m state --state INVALID -j DROP")
    exec("iptables -A OUTPUT -m state --state ESTABLISHED -j ACCEPT")

    # Allow Tor process output
    # 允许特定用户直连外网。
    exec(
        f"iptables -A OUTPUT -o {_out_if} -m owner --uid-owner {_tor_uid} -p tcp -m tcp --tcp-flags FIN,SYN,RST,ACK SYN -m state --state NEW -j ACCEPT"
    )
    exec(
        f"iptables -A OUTPUT -o {_out_if} -m owner --uid-owner {_nottor_uid} -p tcp -m tcp --tcp-flags FIN,SYN,RST,ACK SYN -m state --state NEW -j ACCEPT"
    )

    # Allow loopback output
    exec("iptables -A OUTPUT -d 127.0.0.1/32 -o lo -j ACCEPT")
    # Tor transproxy magic
    exec(
        f"iptables -A OUTPUT -d 127.0.0.1/32 -p tcp -m tcp --dport {_trans_port} --tcp-flags FIN,SYN,RST,ACK SYN -j ACCEPT"
    )

    # Allow OUTPUT to lan hosts in $_non_tor
    for _lan in _non_tor.split(" "):
        exec(f"iptables -A OUTPUT -d {_lan} -j ACCEPT")

    exec("iptables -A OUTPUT -j DROP")
    # -P 参数表示默认策略
    # ### Set default policies to DROP
    exec("iptables -P INPUT DROP")
    exec("iptables -P FORWARD DROP")
    exec("iptables -P OUTPUT DROP")

    # # Set default policies to DROP for IPv6
    # exec("ip6tables -P INPUT DROP")
    # exec("ip6tables -P FORWARD DROP")
    # exec("ip6tables -P OUTPUT DROP")

    # ##
    # # vpn 内网透明代理
    # ##
    # exec("iptables -t nat -A PREROUTING -s 10.12.12.0/24 -p udp --dport 53 -j REDIRECT --to-ports 5353")
    # # exec("iptables -t nat -A PREROUTING -s 10.12.12.0/24 -p udp --dport 5353 -j REDIRECT --to-ports 5353")

    # exec("iptables -t nat -A PREROUTING -s 10.12.12.0/24 -p tcp --syn -j REDIRECT --to-ports {_trans_port}".format(
    #     _trans_port=_trans_port
    # ))

    # #
    # # 下面这三行是根据tun0接口来做透明代理转发，有效的。
    # #
    exec(
        "iptables -t nat -A PREROUTING -i tun0 -p udp --dport 53 -j REDIRECT --to-ports 5353"
    )
    exec(
        "iptables -t nat -A PREROUTING -i tun0 -p udp --dport 5353 -j REDIRECT --to-ports 5353"
    )
    exec(
        "iptables -t nat -A PREROUTING -i tun0 -p tcp --syn -j REDIRECT --to-ports 9040"
    )



# coding: utf-8

"""
    Mtmai API

    The Mtmai API

    The version of the OpenAPI document: 1.0.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from pydantic import BaseModel, ConfigDict, StrictStr, field_validator
from typing import Any, ClassVar, Dict, List, Optional
from mtmai.gomtmclients.rest.models.agent_config import AgentConfig
from mtmai.gomtmclients.rest.models.model_config import ModelConfig
from mtmai.gomtmclients.rest.models.termination_config import TerminationConfig
from typing import Optional, Set
from typing_extensions import Self

class SelectorGroupChatConfig(BaseModel):
    """
    SelectorGroupChatConfig
    """ # noqa: E501
    component_type: StrictStr
    version: Optional[StrictStr] = None
    description: Optional[StrictStr] = None
    name: Optional[StrictStr] = None
    participants: Optional[List[AgentConfig]] = None
    team_type: Optional[StrictStr] = None
    termination_condition: Optional[TerminationConfig] = None
    selector_prompt: Optional[StrictStr] = None
    model_client: Optional[ModelConfig] = None
    __properties: ClassVar[List[str]] = ["component_type", "version", "description", "name", "participants", "team_type", "termination_condition", "selector_prompt", "model_client"]

    @field_validator('team_type')
    def team_type_validate_enum(cls, value):
        """Validates the enum"""
        if value is None:
            return value

        if value not in set(['SelectorGroupChat']):
            raise ValueError("must be one of enum values ('SelectorGroupChat')")
        return value

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of SelectorGroupChatConfig from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([
        ])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of each item in participants (list)
        _items = []
        if self.participants:
            for _item_participants in self.participants:
                if _item_participants:
                    _items.append(_item_participants.to_dict())
            _dict['participants'] = _items
        # override the default output from pydantic by calling `to_dict()` of termination_condition
        if self.termination_condition:
            _dict['termination_condition'] = self.termination_condition.to_dict()
        # override the default output from pydantic by calling `to_dict()` of model_client
        if self.model_client:
            _dict['model_client'] = self.model_client.to_dict()
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of SelectorGroupChatConfig from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "component_type": obj.get("component_type"),
            "version": obj.get("version"),
            "description": obj.get("description"),
            "name": obj.get("name"),
            "participants": [AgentConfig.from_dict(_item) for _item in obj["participants"]] if obj.get("participants") is not None else None,
            "team_type": obj.get("team_type"),
            "termination_condition": TerminationConfig.from_dict(obj["termination_condition"]) if obj.get("termination_condition") is not None else None,
            "selector_prompt": obj.get("selector_prompt"),
            "model_client": ModelConfig.from_dict(obj["model_client"]) if obj.get("model_client") is not None else None
        })
        return _obj





from langchain_core.messages import HumanMessage
from langchain_core.runnables import Runnable, RunnableConfig
from langchain_core.tools import BaseTool

from mtmai.core.logging import get_logger
from mtmai.models.graph_config import HomeChatState
logger = get_logger()

class ShowAdminPanelTool(BaseTool):
    name: str = "ShowAdminPanelTool"
    description: str = (
        r"当用户明确要求显示管理面板、后台面板时,显示管理面板给用户进行下一步的操作"
    )
    response_format: str = "content_and_artifact"

    def _run(self):
        return (
            "Operation successful",
            {
                "artiface_type": "AdminView",
                "title": "管理面板",
                "props": {
                    "title": "管理面板",
                },
            },
        )


class ShowTasksInputPanelTool(BaseTool):
    name: str = "ShowTasksInputPanel"
    description: str = (
        r"有用的工具, 向用户UI 显示任务输入的选择器, 让用户选择需要运行的任务"
    )
    response_format: str = "content_and_artifact"

    def _run(self):
        return (
            "Operation successful",
            {
                "artiface_type": "TaskSelect",
                "title": "任务选择器",
                "props": {
                    "title": "任务选择器",
                },
            },
        )


show_admin_panel_tool = ShowAdminPanelTool()
form_tool = ShowTasksInputPanelTool()
human_node_tools = [
    show_admin_panel_tool,
    form_tool,
]


class HumanNode:
    def __init__(self, runnable: Runnable):
        self.runnable = runnable

    async def __call__(self, state: HomeChatState, config: RunnableConfig):
        logger.info("进入 human_node")
        messages = state.messages
        user_input = state.user_input
        # if user_input:
        #     messages.append(HumanMessage(content=user_input))

        if user_input == "/1":
            logger.info("特殊指令1")

        return {
            "messages": HumanMessage(content=user_input),
        }



# coding: utf-8

"""
    Mtmai API

    The Mtmai API

    The version of the OpenAPI document: 1.0.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from pydantic import BaseModel, ConfigDict, Field, StrictFloat, StrictInt, StrictStr
from typing import Any, ClassVar, Dict, List, Optional, Union
from mtmai.gomtmclients.rest.models.api_resource_meta import APIResourceMeta
from mtmai.gomtmclients.rest.models.chat_message import ChatMessage
from mtmai.gomtmclients.rest.models.llm_config import LlmConfig
from typing import Optional, Set
from typing_extensions import Self

class BlogTaskState(BaseModel):
    """
    BlogTaskState
    """ # noqa: E501
    metadata: APIResourceMeta
    thread_id: Optional[StrictStr] = Field(default=None, description="线程ID", alias="threadId")
    messages: List[ChatMessage] = Field(description="聊天消息")
    llm: Optional[LlmConfig] = None
    prompt: Optional[StrictStr] = Field(default=None, description="关键提示语")
    blog_description: StrictStr = Field(description="博客站点功能定位描述", alias="blogDescription")
    blog_keywords: List[StrictStr] = Field(description="博客的SEO关键字", alias="blogKeywords")
    cur_topic_to_gen: Optional[StrictStr] = Field(default=None, description="当前生成文章使用的主题", alias="curTopicToGen")
    cur_article_state: Optional[GenArticleState] = Field(default=None, description="当前正在生成的文章", alias="curArticleState")
    step_description: Optional[StrictStr] = Field(default=None, description="当前步骤描述", alias="stepDescription")
    running_state: Optional[StrictStr] = Field(default=None, description="运行状态", alias="runningState")
    day_publishd_count: Union[StrictFloat, StrictInt] = Field(description="已经完成的日更天子数量", alias="dayPublishdCount")
    day_publish_count_hint: Union[StrictFloat, StrictInt] = Field(description="建议日更数", alias="dayPublishCountHint")
    __properties: ClassVar[List[str]] = ["metadata", "threadId", "messages", "llm", "prompt", "blogDescription", "blogKeywords", "curTopicToGen", "curArticleState", "stepDescription", "runningState", "dayPublishdCount", "dayPublishCountHint"]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of BlogTaskState from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([
        ])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of metadata
        if self.metadata:
            _dict['metadata'] = self.metadata.to_dict()
        # override the default output from pydantic by calling `to_dict()` of each item in messages (list)
        _items = []
        if self.messages:
            for _item_messages in self.messages:
                if _item_messages:
                    _items.append(_item_messages.to_dict())
            _dict['messages'] = _items
        # override the default output from pydantic by calling `to_dict()` of llm
        if self.llm:
            _dict['llm'] = self.llm.to_dict()
        # override the default output from pydantic by calling `to_dict()` of cur_article_state
        if self.cur_article_state:
            _dict['curArticleState'] = self.cur_article_state.to_dict()
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of BlogTaskState from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "metadata": APIResourceMeta.from_dict(obj["metadata"]) if obj.get("metadata") is not None else None,
            "threadId": obj.get("threadId"),
            "messages": [ChatMessage.from_dict(_item) for _item in obj["messages"]] if obj.get("messages") is not None else None,
            "llm": LlmConfig.from_dict(obj["llm"]) if obj.get("llm") is not None else None,
            "prompt": obj.get("prompt"),
            "blogDescription": obj.get("blogDescription"),
            "blogKeywords": obj.get("blogKeywords"),
            "curTopicToGen": obj.get("curTopicToGen"),
            "curArticleState": GenArticleState.from_dict(obj["curArticleState"]) if obj.get("curArticleState") is not None else None,
            "stepDescription": obj.get("stepDescription"),
            "runningState": obj.get("runningState"),
            "dayPublishdCount": obj.get("dayPublishdCount") if obj.get("dayPublishdCount") is not None else 0,
            "dayPublishCountHint": obj.get("dayPublishCountHint") if obj.get("dayPublishCountHint") is not None else 10
        })
        return _obj

from mtmai.gomtmclients.rest.models.gen_article_state import GenArticleState
# TODO: Rewrite to not use raise_errors
BlogTaskState.model_rebuild(raise_errors=False)




import uuid
from datetime import datetime

from pgvector.sqlalchemy import Vector
from sqlalchemy.dialects.postgresql import TSVECTOR
from sqlmodel import JSON, Column, Field, SQLModel


class SearchIndexBase(SQLModel):
    id: uuid.UUID = Field(default_factory=uuid.uuid4, primary_key=True)
    title: str = Field(...)
    content_type: str = Field(index=True)  # 例如 'post', 'site', 'thread'
    content_id: uuid.UUID = Field(...)
    description: str | None = Field(default=None)  # 显示在列表 item 中的描述栏
    created_at: datetime = Field(default_factory=datetime.now, nullable=False)
    updated_at: datetime = Field(default_factory=datetime.now, nullable=False)


class SearchIndex(SearchIndexBase, table=True):
    __tablename__ = "search_index"
    workspace_id: uuid.UUID | None = Field(default=None, index=True)
    owner_id: uuid.UUID | None = Field(
        foreign_key="user.id", index=True, nullable=False, ondelete="CASCADE"
    )
    is_public: bool = Field(default=True, index=True)
    meta: dict | None = Field(default={}, sa_column=Column(JSON))  # 用于存储额外的元数
    search_vector: str = Field(sa_column=Column(TSVECTOR))
    embedding: list[float] = Field(sa_column=Column(Vector()))
    expired_at: datetime | None = Field(default=None)
    content_summary: str = Field(
        index=True, description="Content summary for full-text search"
    )
    is_deleted: bool = Field(default=False, index=True)


class SearchIndexPublic(SearchIndexBase):
    pass


class SearchIndexResponse(SQLModel):
    data: list[SearchIndexPublic]
    count: int


# Deprecated: Use ListviewRequest instead
class SearchRequest(SQLModel):
    dataType: str | None = None
    q: str | None = None  # 搜索关键词
    limit: int = 100
    skip: int = 0
    # 搜索来自什么应用
    app: str | None = None



import os
from urllib.parse import urlparse

from ..utils.env import is_in_docker, is_in_testing, is_in_vercel


def is_in_dev() -> bool:
    """是否处于开发环境"""
    from mtmai.core.config import settings

    if is_in_gitpod():
        return True
    return (
        not is_in_docker()
        and not is_in_vercel()
        and not is_in_testing()
        and not settings.is_production
    )


def backend_url_base() -> str:
    from mtmai.core.config import settings

    if settings.Serve_ADDR:
        return f"http://{settings.Serve_ADDR}"

    gitpod_workspace_url = os.environ.get("GITPOD_WORKSPACE_URL")

    if gitpod_workspace_url:
        uri1 = urlparse(gitpod_workspace_url)
        return f"https://{settings.PORT}-{uri1.hostname}"
    return f"http://localhost:{settings.PORT}"


def get_server_host():
    gitpod_workspace_url = os.environ.get("GITPOD_WORKSPACE_URL")
    if gitpod_workspace_url:
        uri1 = urlparse(gitpod_workspace_url)
        return uri1.hostname
    return "0.0.0.0"  # noqa: S104


# def abs_url(req: Request, path_name: str = ""):
#     """
#     获取完整后端url, (自动处理反代的情况)
#     """
#     x_forwardd_host = req.headers.get("x-forwarded-host")
#     logger = getLogger()
#     logger.info("获取绝对网址, headers:", req.headers)
#     base_url = backend_url_base()
#     if x_forwardd_host:
#         # x_forwardd_port = req.headers.get("x-forwarded-port")
#         x_forwardd_proto = req.headers.get("x-forwarded-proto")
#         base_url = f"{x_forwardd_proto}://{x_forwardd_host}"

#     return f"{base_url}{path_name}"


def is_in_gitpod() -> bool:
    return os.getenv("GITPOD_WORKSPACE_URL")



from typing import Unpack

from autogen_core.models import ModelFamily
from autogen_ext.models.openai import OpenAIChatCompletionClient
from autogen_ext.models.openai.config import (
    OpenAIClientConfiguration,
    OpenAIClientConfigurationConfigModel,
)

# from json_repair import repair_json
# from loguru import logger


class MtmOpenAIChatCompletionClient(OpenAIChatCompletionClient):
    """
    官方的 OpenAIChatCompletionClient 有一个问题在于序列化为 component 时,
    会丢失 model_info 信息,因此我们重写了这个类,并添加了 model_info 信息
    """

    def __init__(self, **kwargs: Unpack[OpenAIClientConfiguration]):
        super().__init__(**kwargs)
        self.model_info_to_keep = kwargs["model_info"]

    def _to_config(self) -> OpenAIClientConfigurationConfigModel:
        copied_config = self._raw_config.copy()
        copied_config["model_info"] = self.model_info_to_keep
        return OpenAIClientConfigurationConfigModel(**copied_config)


# class LoggingModelClient:
#     """
#     日志记录模型客户端
#     """

#     def __init__(self, wrapped_client):
#         self.wrapped_client = wrapped_client

#     async def create(self, *args: Any, **kwargs: Any) -> Any:
#         try:
#             response = await self.wrapped_client.create(*args, **kwargs)
#             if kwargs.get("json_output", False):
#                 # 修正json格式
#                 if isinstance(response.content, str):
#                     response.content = repair_json(response.content)

#             logger.info(
#                 "OpenAI API Response",
#                 request_args=args,
#                 request_kwargs=kwargs,
#                 response_content=response.content,
#             )
#             return response
#         except Exception as e:
#             logger.exception(
#                 "OpenAI API Error", error=str(e), error_type=type(e).__name__
#             )
#             raise


def get_oai_Model():
    model_client = MtmOpenAIChatCompletionClient(
        model="deepseek-r1:1.5b",
        base_url="https://llama3-3-70b.lepton.run/api/v1/",
        api_key="YLJU3oah5ZwNv1HzAGOeVwfvDfUWB6yb",
        model_info={
            "vision": False,
            "function_calling": True,
            "json_output": True,
            "family": ModelFamily.R1,
        },
    )
    return model_client
    return model_client



#!/usr/bin/env python3

"""

         888                                                  888         d8b
         888                                                  888         Y8P
         888                                                  888
 .d8888b 88888b.  888d888 .d88b.  88888b.d88b.   .d88b.   .d88888 888d888 888 888  888  .d88b.  888d888
d88P"    888 "88b 888P"  d88""88b 888 "888 "88b d8P  Y8b d88" 888 888P"   888 888  888 d8P  Y8b 888P"
888      888  888 888    888  888 888  888  888 88888888 888  888 888     888 Y88  88P 88888888 888
Y88b.    888  888 888    Y88..88P 888  888  888 Y8b.     Y88b 888 888     888  Y8bd8P  Y8b.     888
 "Y8888P 888  888 888     "Y88P"  888  888  888  "Y8888   "Y88888 888     888   Y88P    "Y8888  888   88888888

by UltrafunkAmsterdam (https://github.com/ultrafunkamsterdam)

"""
from __future__ import annotations


__version__ = "3.4.7"

import json
import logging
import os
import re
import shutil
import subprocess
import sys
import tempfile
import time
from weakref import finalize

import selenium.webdriver.chrome.service
import selenium.webdriver.chrome.webdriver
from selenium.webdriver.common.by import By
import selenium.webdriver.common.service
import selenium.webdriver.remote.command
import selenium.webdriver.remote.webdriver

from .cdp import CDP
from .dprocess import start_detached
from .options import ChromeOptions
from .patcher import IS_POSIX
from .patcher import Patcher
from .reactor import Reactor
from .webelement import UCWebElement
from .webelement import WebElement


__all__ = (
    "Chrome",
    "ChromeOptions",
    "Patcher",
    "Reactor",
    "CDP",
    "find_chrome_executable",
)

logger = logging.getLogger("uc")
logger.setLevel(logging.getLogger().getEffectiveLevel())


class Chrome(selenium.webdriver.chrome.webdriver.WebDriver):
    """

    Controls the ChromeDriver and allows you to drive the browser.

    The webdriver file will be downloaded by this module automatically,
    you do not need to specify this. however, you may if you wish.

    Attributes
    ----------

    Methods
    -------

    reconnect()

        this can be useful in case of heavy detection methods
        -stops the chromedriver service which runs in the background
        -starts the chromedriver service which runs in the background
        -recreate session


    start_session(capabilities=None, browser_profile=None)

        differentiates from the regular method in that it does not
        require a capabilities argument. The capabilities are automatically
        recreated from the options at creation time.

    --------------------------------------------------------------------------
        NOTE:
            Chrome has everything included to work out of the box.
            it does not `need` customizations.
            any customizations MAY lead to trigger bot migitation systems.

    --------------------------------------------------------------------------
    """

    _instances = set()
    session_id = None
    debug = False

    def __init__(
        self,
        options=None,
        user_data_dir=None,
        driver_executable_path=None,
        browser_executable_path=None,
        port=0,
        enable_cdp_events=False,
        service_args=None,
        service_creationflags=None,
        desired_capabilities=None,
        advanced_elements=False,
        service_log_path=None,
        keep_alive=True,
        log_level=0,
        headless=False,
        version_main=None,
        patcher_force_close=False,
        suppress_welcome=True,
        use_subprocess=True,
        debug=False,
        no_sandbox=True,
        user_multi_procs: bool = False,
        **kw,
    ):
        """
        Creates a new instance of the chrome driver.

        Starts the service and then creates new instance of chrome driver.

        Parameters
        ----------

        options: ChromeOptions, optional, default: None - automatic useful defaults
            this takes an instance of ChromeOptions, mainly to customize browser behavior.
            anything other dan the default, for example extensions or startup options
            are not supported in case of failure, and can probably lowers your undetectability.


        user_data_dir: str , optional, default: None (creates temp profile)
            if user_data_dir is a path to a valid chrome profile directory, use it,
            and turn off automatic removal mechanism at exit.

        driver_executable_path: str, optional, default: None(=downloads and patches new binary)

        browser_executable_path: str, optional, default: None - use find_chrome_executable
            Path to the browser executable.
            If not specified, make sure the executable's folder is in $PATH

        port: int, optional, default: 0
            port to be used by the chromedriver executable, this is NOT the debugger port.
            leave it at 0 unless you know what you are doing.
            the default value of 0 automatically picks an available port.

        enable_cdp_events: bool, default: False
            :: currently for chrome only
            this enables the handling of wire messages
            when enabled, you can subscribe to CDP events by using:

                driver.add_cdp_listener("Network.dataReceived", yourcallback)
                # yourcallback is an callable which accepts exactly 1 dict as parameter


        service_args: list of str, optional, default: None
            arguments to pass to the driver service

        desired_capabilities: dict, optional, default: None - auto from config
            Dictionary object with non-browser specific capabilities only, such as "item" or "loggingPref".

        advanced_elements:  bool, optional, default: False
            makes it easier to recognize elements like you know them from html/browser inspection, especially when working
            in an interactive environment

            default webelement repr:
            <selenium.webdriver.remote.webelement.WebElement (session="85ff0f671512fa535630e71ee951b1f2", element="6357cb55-92c3-4c0f-9416-b174f9c1b8c4")>

            advanced webelement repr
            <WebElement(<a class="mobile-show-inline-block mc-update-infos init-ok" href="#" id="main-cat-switcher-mobile">)>

            note: when retrieving large amounts of elements ( example: find_elements_by_tag("*") ) and print them, it does take a little more time.


        service_log_path: str, optional, default: None
             path to log information from the driver.

        keep_alive: bool, optional, default: True
             Whether to configure ChromeRemoteConnection to use HTTP keep-alive.

        log_level: int, optional, default: adapts to python global log level

        headless: bool, optional, default: False
            can also be specified in the options instance.
            Specify whether you want to use the browser in headless mode.
            warning: this lowers undetectability and not fully supported.

        version_main: int, optional, default: None (=auto)
            if you, for god knows whatever reason, use
            an older version of Chrome. You can specify it's full rounded version number
            here. Example: 87 for all versions of 87

        patcher_force_close: bool, optional, default: False
            instructs the patcher to do whatever it can to access the chromedriver binary
            if the file is locked, it will force shutdown all instances.
            setting it is not recommended, unless you know the implications and think
            you might need it.

        suppress_welcome: bool, optional , default: True
            a "welcome" alert might show up on *nix-like systems asking whether you want to set
            chrome as your default browser, and if you want to send even more data to google.
            now, in case you are nag-fetishist, or a diagnostics data feeder to google, you can set this to False.
            Note: if you don't handle the nag screen in time, the browser loses it's connection and throws an Exception.

        use_subprocess: bool, optional , default: True,

            False (the default) makes sure Chrome will get it's own process (so no subprocess of chromedriver.exe or python
                This fixes a LOT of issues, like multithreaded run, but mst importantly. shutting corectly after
                program exits or using .quit()
                you should be knowing what you're doing, and know how python works.

              unfortunately, there  is always an edge case in which one would like to write an single script with the only contents being:
              --start script--
              import undetected_chromedriver as uc
              d = uc.Chrome()
              d.get('https://somesite/')
              ---end script --

              and will be greeted with an error, since the program exists before chrome has a change to launch.
              in that case you can set this to `True`. The browser will start via subprocess, and will keep running most of times.
              ! setting it to True comes with NO support when being detected. !

        no_sandbox: bool, optional, default=True
             uses the --no-sandbox option, and additionally does suppress the "unsecure option" status bar
             this option has a default of True since many people seem to run this as root (....) , and chrome does not start
             when running as root without using --no-sandbox flag.

        user_multi_procs:
            set to true when you are using multithreads/multiprocessing
            ensures not all processes are trying to modify a binary which is in use by another.
            for this to work. YOU MUST HAVE AT LEAST 1 UNDETECTED_CHROMEDRIVER BINARY IN YOUR ROAMING DATA FOLDER.
            this requirement can be easily satisfied, by just running this program "normal" and close/kill it.


        """

        finalize(self, self._ensure_close, self)
        self.debug = debug
        self.patcher = Patcher(
            executable_path=driver_executable_path,
            force=patcher_force_close,
            version_main=version_main,
            user_multi_procs=user_multi_procs,
        )
        # self.patcher.auto(user_multiprocess = user_multi_num_procs)
        chrome_version = self.patcher.auto()

        # self.patcher = patcher
        if not options:
            options = ChromeOptions()

        try:
            if hasattr(options, "_session") and options._session is not None:
                #  prevent reuse of options,
                #  as it just appends arguments, not replace them
                #  you'll get conflicts starting chrome
                raise RuntimeError("you cannot reuse the ChromeOptions object")
        except AttributeError:
            pass

        options._session = self

        if not options.debugger_address:
            debug_port = (
                port
                if port != 0
                else selenium.webdriver.common.service.utils.free_port()
            )
            debug_host = "127.0.0.1"
            options.debugger_address = "%s:%d" % (debug_host, debug_port)
        else:
            debug_host, debug_port = options.debugger_address.split(":")
            debug_port = int(debug_port)

        if enable_cdp_events:
            options.set_capability(
                "goog:loggingPrefs", {"performance": "ALL", "browser": "ALL"}
            )

        options.add_argument("--remote-debugging-host=%s" % debug_host)
        options.add_argument("--remote-debugging-port=%s" % debug_port)

        if user_data_dir:
            options.add_argument("--user-data-dir=%s" % user_data_dir)

        language, keep_user_data_dir = None, bool(user_data_dir)

        # see if a custom user profile is specified in options
        for arg in options.arguments:

            if any([_ in arg for _ in ("--headless", "headless")]):
                options.arguments.remove(arg)
                options.headless = True

            if "lang" in arg:
                m = re.search("(?:--)?lang(?:[ =])?(.*)", arg)
                try:
                    language = m[1]
                except IndexError:
                    logger.debug("will set the language to en-US,en;q=0.9")
                    language = "en-US,en;q=0.9"

            if "user-data-dir" in arg:
                m = re.search("(?:--)?user-data-dir(?:[ =])?(.*)", arg)
                try:
                    user_data_dir = m[1]
                    logger.debug(
                        "user-data-dir found in user argument %s => %s" % (arg, m[1])
                    )
                    keep_user_data_dir = True

                except IndexError:
                    logger.debug(
                        "no user data dir could be extracted from supplied argument %s "
                        % arg
                    )

        if not user_data_dir:
            # backward compatiblity
            # check if an old uc.ChromeOptions is used, and extract the user data dir

            if hasattr(options, "user_data_dir") and getattr(
                options, "user_data_dir", None
            ):
                import warnings

                warnings.warn(
                    "using ChromeOptions.user_data_dir might stop working in future versions."
                    "use uc.Chrome(user_data_dir='/xyz/some/data') in case you need existing profile folder"
                )
                options.add_argument("--user-data-dir=%s" % options.user_data_dir)
                keep_user_data_dir = True
                logger.debug(
                    "user_data_dir property found in options object: %s" % user_data_dir
                )

            else:
                user_data_dir = os.path.normpath(tempfile.mkdtemp())
                keep_user_data_dir = False
                arg = "--user-data-dir=%s" % user_data_dir
                options.add_argument(arg)
                logger.debug(
                    "created a temporary folder in which the user-data (profile) will be stored during this\n"
                    "session, and added it to chrome startup arguments: %s" % arg
                )

        if not language:
            try:
                import locale

                language = locale.getdefaultlocale()[0].replace("_", "-")
            except Exception:
                pass
            if not language:
                language = "en-US"

        options.add_argument("--lang=%s" % language)

        if not options.binary_location:
            options.binary_location = (
                browser_executable_path or find_chrome_executable(chrome_version)
            )
        if not os.path.exists(options.binary_location):
            time.sleep(5)
            # 如果没有安装，可以在下面的链接下载安装：https://www.google.com/chrome/beta/
            print(f"""\n\n\n要想过Cloudflare验证，需要以下目录存在115版本的Chrome Beta版浏览器，注意是Beta版不是正式版：C:\Program Files\Google\Chrome Beta
                    如果Beta版本不是115，请在软件下载目录中找到Chrome_Beta_115_win64.7z压缩包，然后解压并复制（覆盖）为C:\Program Files\Google\Chrome Beta目录即可。

                    请手动关闭此程序，配置完成后重新执行任务。

                    """)
            print("""To pass the Cloudflare verification, you need the following directory to exist in the 115 version of Chrome Beta, note that it is the Beta version not the official version: C:\Program Files\Google\Chrome Beta,
                    If the Beta version is not 115, please find the Chrome_Beta_115_win64.7z compressed package in the software download directory, then unzip and copy (overwrite) to the C:\Program Files\Google\Chrome Beta directory.

                    Please close this program manually and re-execute the task after the configuration is complete.

                    """)

            time.sleep(100)
        else:
            folder_path = os.path.dirname(os.path.abspath(options.binary_location))
            folder_list = [f for f in os.listdir(folder_path) if os.path.isdir(os.path.join(folder_path, f))]
            numeric_folders = [f for f in folder_list if f[0].isdigit()]
            version = numeric_folders[0].split('.')[0]
            if version != "115":
                time.sleep(5)
                print("Chrome Beta版本不是115，请将Chrome Beta的版本替换为115， 方法为下载115版本的Chrome Beta浏览器，然后解压并覆盖C:\Program Files\Google\Chrome Beta目录即可，软件下载目录中有Chrome_Beta_115_win64.7z版本的压缩包，可直接下载后解压替换。")
                print("Chrome Beta version is not 115, please replace the version of Chrome Beta with 115, the method is to download the 115 version of Chrome Beta browser, then unzip and overwrite the C:\Program Files\Google\Chrome Beta directory, the software download directory has Chrome_Beta_115_win64.7z version of the compressed package, you can download and unzip directly to replace.")
                print("\n请手动关闭此程序。\n")
                print("\nPlease close this program manually.\n")
                time.sleep(100)



        print("Options Binary Location: ", options.binary_location)

        self._delay = 3

        self.user_data_dir = user_data_dir
        self.keep_user_data_dir = keep_user_data_dir

        if suppress_welcome:
            options.arguments.extend(["--no-default-browser-check", "--no-first-run"])
        if no_sandbox:
            options.arguments.extend(["--no-sandbox", "--test-type"])

        if headless or options.headless:
            if self.patcher.version_main < 108:
                options.add_argument("--headless=chrome")
            elif self.patcher.version_main >= 108:
                options.add_argument("--headless=new")

        # options.add_argument("--window-size=1920,1080")
        options.add_argument("--start-maximized")
        options.add_argument("--no-sandbox")
        # fixes "could not connect to chrome" error when running
        # on linux using privileged user like root (which i don't recommend)

        options.add_argument(
            "--log-level=%d" % log_level
            or divmod(logging.getLogger().getEffectiveLevel(), 10)[0]
        )

        if hasattr(options, "handle_prefs"):
            options.handle_prefs(user_data_dir)

        # fix exit_type flag to prevent tab-restore nag
        try:
            with open(
                os.path.join(user_data_dir, "Default/Preferences"),
                encoding="latin1",
                mode="r+",
            ) as fs:
                config = json.load(fs)
                if config["profile"]["exit_type"] is not None:
                    # fixing the restore-tabs-nag
                    config["profile"]["exit_type"] = None
                fs.seek(0, 0)
                json.dump(config, fs)
                fs.truncate()  # the file might be shorter
                logger.debug("fixed exit_type flag")
        except Exception as e:
            logger.debug("did not find a bad exit_type flag ")

        self.options = options

        if not desired_capabilities:
            desired_capabilities = options.to_capabilities()

        if not use_subprocess:
            self.browser_pid = start_detached(
                options.binary_location, *options.arguments
            )
        else:
            browser = subprocess.Popen(
                [options.binary_location, *options.arguments],
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                close_fds=IS_POSIX,
            )
            self.browser_pid = browser.pid

        if service_creationflags:
            service = selenium.webdriver.common.service.Service(
                self.patcher.executable_path, port, service_args, service_log_path
            )
            for attr_name in ("creationflags", "creation_flags"):
                if hasattr(service, attr_name):
                    setattr(service, attr_name, service_creationflags)
                    break
        else:
            service = None

        super(Chrome, self).__init__(
            executable_path=self.patcher.executable_path,
            port=port,
            options=options,
            service_args=service_args,
            desired_capabilities=desired_capabilities,
            service_log_path=service_log_path,
            keep_alive=keep_alive,
            service=service,  # needed or the service will be re-created
        )

        self.reactor = None

        if enable_cdp_events:
            if logging.getLogger().getEffectiveLevel() == logging.DEBUG:
                logging.getLogger(
                    "selenium.webdriver.remote.remote_connection"
                ).setLevel(20)
            reactor = Reactor(self)
            reactor.start()
            self.reactor = reactor

        if advanced_elements:
            self._web_element_cls = UCWebElement
        else:
            self._web_element_cls = WebElement

        if options.headless:
            self._configure_headless()

    def _configure_headless(self):
        orig_get = self.get
        logger.info("setting properties for headless")

        def get_wrapped(*args, **kwargs):
            if self.execute_script("return navigator.webdriver"):
                logger.info("patch navigator.webdriver")
                self.execute_cdp_cmd(
                    "Page.addScriptToEvaluateOnNewDocument",
                    {
                        "source": """

                           Object.defineProperty(window, "navigator", {
                                Object.defineProperty(window, "navigator", {
                                  value: new Proxy(navigator, {
                                    has: (target, key) => (key === "webdriver" ? false : key in target),
                                    get: (target, key) =>
                                      key === "webdriver"
                                        ? false
                                        : typeof target[key] === "function"
                                        ? target[key].bind(target)
                                        : target[key],
                                  }),
                                });
                    """
                    },
                )

                logger.info("patch user-agent string")
                self.execute_cdp_cmd(
                    "Network.setUserAgentOverride",
                    {
                        "userAgent": self.execute_script(
                            "return navigator.userAgent"
                        ).replace("Headless", "")
                    },
                )
                self.execute_cdp_cmd(
                    "Page.addScriptToEvaluateOnNewDocument",
                    {
                        "source": """
                            Object.defineProperty(navigator, 'maxTouchPoints', {get: () => 1});
                            Object.defineProperty(navigator.connection, 'rtt', {get: () => 100});

                            // https://github.com/microlinkhq/browserless/blob/master/packages/goto/src/evasions/chrome-runtime.js
                            window.chrome = {
                                app: {
                                    isInstalled: false,
                                    InstallState: {
                                        DISABLED: 'disabled',
                                        INSTALLED: 'installed',
                                        NOT_INSTALLED: 'not_installed'
                                    },
                                    RunningState: {
                                        CANNOT_RUN: 'cannot_run',
                                        READY_TO_RUN: 'ready_to_run',
                                        RUNNING: 'running'
                                    }
                                },
                                runtime: {
                                    OnInstalledReason: {
                                        CHROME_UPDATE: 'chrome_update',
                                        INSTALL: 'install',
                                        SHARED_MODULE_UPDATE: 'shared_module_update',
                                        UPDATE: 'update'
                                    },
                                    OnRestartRequiredReason: {
                                        APP_UPDATE: 'app_update',
                                        OS_UPDATE: 'os_update',
                                        PERIODIC: 'periodic'
                                    },
                                    PlatformArch: {
                                        ARM: 'arm',
                                        ARM64: 'arm64',
                                        MIPS: 'mips',
                                        MIPS64: 'mips64',
                                        X86_32: 'x86-32',
                                        X86_64: 'x86-64'
                                    },
                                    PlatformNaclArch: {
                                        ARM: 'arm',
                                        MIPS: 'mips',
                                        MIPS64: 'mips64',
                                        X86_32: 'x86-32',
                                        X86_64: 'x86-64'
                                    },
                                    PlatformOs: {
                                        ANDROID: 'android',
                                        CROS: 'cros',
                                        LINUX: 'linux',
                                        MAC: 'mac',
                                        OPENBSD: 'openbsd',
                                        WIN: 'win'
                                    },
                                    RequestUpdateCheckStatus: {
                                        NO_UPDATE: 'no_update',
                                        THROTTLED: 'throttled',
                                        UPDATE_AVAILABLE: 'update_available'
                                    }
                                }
                            }

                            // https://github.com/microlinkhq/browserless/blob/master/packages/goto/src/evasions/navigator-permissions.js
                            if (!window.Notification) {
                                window.Notification = {
                                    permission: 'denied'
                                }
                            }

                            const originalQuery = window.navigator.permissions.query
                            window.navigator.permissions.__proto__.query = parameters =>
                                parameters.name === 'notifications'
                                    ? Promise.resolve({ state: window.Notification.permission })
                                    : originalQuery(parameters)

                            const oldCall = Function.prototype.call
                            function call() {
                                return oldCall.apply(this, arguments)
                            }
                            Function.prototype.call = call

                            const nativeToStringFunctionString = Error.toString().replace(/Error/g, 'toString')
                            const oldToString = Function.prototype.toString

                            function functionToString() {
                                if (this === window.navigator.permissions.query) {
                                    return 'function query() { [native code] }'
                                }
                                if (this === functionToString) {
                                    return nativeToStringFunctionString
                                }
                                return oldCall.call(oldToString, this)
                            }
                            // eslint-disable-next-line
                            Function.prototype.toString = functionToString
                            """
                    },
                )
            return orig_get(*args, **kwargs)

        self.get = get_wrapped

    # def _get_cdc_props(self):
    #     return self.execute_script(
    #         """
    #         let objectToInspect = window,
    #             result = [];
    #         while(objectToInspect !== null)
    #         { result = result.concat(Object.getOwnPropertyNames(objectToInspect));
    #           objectToInspect = Object.getPrototypeOf(objectToInspect); }
    #
    #         return result.filter(i => i.match(/^([a-zA-Z]){27}(Array|Promise|Symbol)$/ig))
    #         """
    #     )
    #
    # def _hook_remove_cdc_props(self):
    #     self.execute_cdp_cmd(
    #         "Page.addScriptToEvaluateOnNewDocument",
    #         {
    #             "source": """
    #                 let objectToInspect = window,
    #                     result = [];
    #                 while(objectToInspect !== null)
    #                 { result = result.concat(Object.getOwnPropertyNames(objectToInspect));
    #                   objectToInspect = Object.getPrototypeOf(objectToInspect); }
    #                 result.forEach(p => p.match(/^([a-zA-Z]){27}(Array|Promise|Symbol)$/ig)
    #                                     &&delete window[p]&&console.log('removed',p))
    #                 """
    #         },
    #     )

    def get(self, url):
        # if self._get_cdc_props():
        #     self._hook_remove_cdc_props()
        return super().get(url)

    def add_cdp_listener(self, event_name, callback):
        if (
            self.reactor
            and self.reactor is not None
            and isinstance(self.reactor, Reactor)
        ):
            self.reactor.add_event_handler(event_name, callback)
            return self.reactor.handlers
        return False

    def clear_cdp_listeners(self):
        if self.reactor and isinstance(self.reactor, Reactor):
            self.reactor.handlers.clear()

    def window_new(self):
        self.execute(
            selenium.webdriver.remote.command.Command.NEW_WINDOW, {"type": "window"}
        )

    def tab_new(self, url: str):
        """
        this opens a url in a new tab.
        apparently, that passes all tests directly!

        Parameters
        ----------
        url

        Returns
        -------

        """
        if not hasattr(self, "cdp"):
            from .cdp import CDP

            cdp = CDP(self.options)
            cdp.tab_new(url)

    def reconnect(self, timeout=0.1):
        try:
            self.service.stop()
        except Exception as e:
            logger.debug(e)
        time.sleep(timeout)
        try:
            self.service.start()
        except Exception as e:
            logger.debug(e)

        try:
            self.start_session()
        except Exception as e:
            logger.debug(e)

    def start_session(self, capabilities=None, browser_profile=None):
        if not capabilities:
            capabilities = self.options.to_capabilities()
        super(selenium.webdriver.chrome.webdriver.WebDriver, self).start_session(
            capabilities, browser_profile
        )
        # super(Chrome, self).start_session(capabilities, browser_profile)

    def quit(self):
        try:
            self.service.process.kill()
            logger.debug("webdriver process ended")
        except (AttributeError, RuntimeError, OSError):
            pass
        try:
            self.reactor.event.set()
            logger.debug("shutting down reactor")
        except AttributeError:
            pass
        try:
            os.kill(self.browser_pid, 15)
            logger.debug("gracefully closed browser")
        except Exception as e:  # noqa
            logger.debug(e, exc_info=True)
        if (
            hasattr(self, "keep_user_data_dir")
            and hasattr(self, "user_data_dir")
            and not self.keep_user_data_dir
        ):
            for _ in range(5):
                try:
                    shutil.rmtree(self.user_data_dir, ignore_errors=False)
                except FileNotFoundError:
                    pass
                except (RuntimeError, OSError, PermissionError) as e:
                    logger.debug(
                        "When removing the temp profile, a %s occured: %s\nretrying..."
                        % (e.__class__.__name__, e)
                    )
                else:
                    logger.debug("successfully removed %s" % self.user_data_dir)
                    break
                time.sleep(0.1)

        # dereference patcher, so patcher can start cleaning up as well.
        # this must come last, otherwise it will throw 'in use' errors
        self.patcher = None

    def __getattribute__(self, item):
        if not super().__getattribute__("debug"):
            return super().__getattribute__(item)
        else:
            import inspect

            original = super().__getattribute__(item)
            if inspect.ismethod(original) and not inspect.isclass(original):

                def newfunc(*args, **kwargs):
                    logger.debug(
                        "calling %s with args %s and kwargs %s\n"
                        % (original.__qualname__, args, kwargs)
                    )
                    return original(*args, **kwargs)

                return newfunc
            return original

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.service.stop()
        time.sleep(self._delay)
        self.service.start()
        self.start_session()

    def __hash__(self):
        return hash(self.options.debugger_address)

    def __dir__(self):
        return object.__dir__(self)

    def __del__(self):
        try:
            self.service.process.kill()
        except:  # noqa
            pass
        self.quit()

    @classmethod
    def _ensure_close(cls, self):
        # needs to be a classmethod so finalize can find the reference
        logger.info("ensuring close")
        if (
            hasattr(self, "service")
            and hasattr(self.service, "process")
            and hasattr(self.service.process, "kill")
        ):
            self.service.process.kill()


def find_chrome_executable(version):
    """
    Finds the chrome, chrome beta, chrome canary, chromium executable

    Returns
    -------
    executable_path :  str
        the full file path to found executable

    """
    candidates = set()
    if IS_POSIX:
        for item in os.environ.get("PATH").split(os.pathsep):
            for subitem in (
                "google-chrome",
                "chromium",
                "chromium-browser",
                "chrome",
                "google-chrome-stable",
            ):
                candidates.add(os.sep.join((item, subitem)))
        if "darwin" in sys.platform:
            candidates.update(
                [
                    "/Applications/Google Chrome.app/Contents/MacOS/Google Chrome",
                    "/Applications/Chromium.app/Contents/MacOS/Chromium",
                ]
            )
    else:
        for item in map(
            os.environ.get,
            ("PROGRAMFILES", "PROGRAMFILES(X86)", "LOCALAPPDATA", "PROGRAMW6432"),
        ):
            if item is not None:
                for subitem in (
                    "Google/Chrome/Application",
                    "Google/Chrome Beta/Application",
                    "Google/Chrome Canary/Application",
                ):
                    candidates.add(os.sep.join((item, subitem, "chrome.exe")))
    for candidate in candidates:
        if os.path.exists(candidate) and os.access(candidate, os.X_OK):
            print(f"""\n\n\n要想过Cloudflare验证，需要满足以下条件：
                  自己的环境已经安装了115版本的Chrome Beta版浏览器，注意是Beta版不是正式版，且浏览器安装路径必须保持不变，在C:\Program Files\Google\Chrome Beta\Application\chrome.exe
                  如果没有安装，可以在下面的链接下载安装：https://www.google.com/chrome/beta/
                  软件将会使用以下目录的Chrome Beta浏览器：", {os.path.normpath(candidate)}, "，请检查此浏览器版本是否为 115 版本的Beta浏览器，如果不是将无法运行。""")
            # print("The software will use the Chrome browser in the following directory:", os.path.normpath(candidate), "Please check if the version of this browser is version " + str(version) + ", if not, it will not be able to run.\n\n\n")
            print(f"""The software will use the Chrome browser in the following directory: {os.path.normpath(candidate)}, Please check if the version of this browser is version 115, if not, it will not be able to run.\n\n\n""")
            time.sleep(5)
            return os.path.normpath(candidate)



from enum import IntEnum, unique


@unique
class WriteMode(IntEnum):
    Create = 0  # 新建模式|Create Mode
    Append = 1  # 追加模式|Append Mode
    MySQL = 2  # MySQL模式|MySQL Mode
    Json = 3  # Json模式|Json Mode


@unique
class DataWriteMode(IntEnum):
    Append = 1  # 追加模式|Append Mode
    Cover = 2  # 覆盖模式|Cover Mode
    Rename = 3  # 重命名模式|Rename Mode


@unique
class GraphOption(IntEnum):
    Get = 1  # 打开网页操作|Open Web
    Click = 2  # 点击操作|Click
    Extract = 3  # 提取数据操作|Extract Data
    Input = 4  # 输入操作|Input
    Custom = 5  # 自定义操作|Custom
    Move = 7  # 移动操作|Move
    Loop = 8  # 循环操作|Loop



# 使用 pytorch 从零开始构建 bert 模型， 不过由于数据量太小看不出明显的预测结果，
# 教程： https://neptune.ai/blog/how-to-code-bert-using-pytorch-tutorial
# notebook: https://colab.research.google.com/drive/13FjI_uXaw8JJGjzjVX3qKSLyW9p3b6OV?usp=sharing

import logging
import math
import re
from random import *  # noqa: F403

import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim

logger = logging.getLogger()

# 训练数据集
text = (
    "Hello, how are you? I am Romeo.\n"
    "Hello, Romeo My name is Juliet. Nice to meet you.\n"
    "Nice meet you too. How are you today?\n"
    "Great. My baseball team won the competition.\n"
    "Oh Congratulations, Juliet\n"
    "Thanks you Romeo"
)


# 获取注意力的填充掩码
def get_attn_pad_mask(seq_q, seq_k):
    batch_size, len_q = seq_q.size()
    batch_size, len_k = seq_k.size()
    # # 等于零是PAD标记
    pad_attn_mask = seq_k.data.eq(0).unsqueeze(
        1
    )  # batch_size x 1 x len_k(=len_q), one is masking
    return pad_attn_mask.expand(batch_size, len_q, len_k)  # batch_size x len_q x len_k


# GELU 激活函数
def gelu(x):
    return x * 0.5 * (1.0 + torch.erf(x / math.sqrt(2.0)))


def train():
    # 处理文本数据
    sentences = re.sub("[.,!?\\-]", "", text.lower()).split(
        "\n"
    )  # filter '.', ',', '?', '!'
    word_list = list(set(" ".join(sentences).split()))
    word_dict = {"[PAD]": 0, "[CLS]": 1, "[SEP]": 2, "[MASK]": 3}

    for i, w in enumerate(word_list):
        word_dict[w] = i + 4
    number_dict = {i: w for i, w in enumerate(word_dict)}
    vocab_size = len(word_dict)

    token_list = list()
    for sentence in sentences:
        arr = [word_dict[s] for s in sentence.split()]
        token_list.append(arr)

    logger.info("show token list")
    print(token_list)

    maxlen = 30  # 序列最大长度
    batch_size = 6
    max_pred = 5  # 最大预测的token数量
    n_layers = 6  # 编码器层数
    n_heads = 12  # 多头注意力头数
    d_model = 768  # 嵌入维度
    d_ff = 768 * 4  # FeedForward维度
    d_k = d_v = 64  # K和V的维度
    n_segments = 2

    def make_batch():
        batch = []
        positive = negative = 0
        while positive != batch_size / 2 or negative != batch_size / 2:
            tokens_a_index, tokens_b_index = (
                randrange(len(sentences)),
                randrange(len(sentences)),
            )
            tokens_a, tokens_b = token_list[tokens_a_index], token_list[tokens_b_index]

            input_ids = (
                [word_dict["[CLS]"]]
                + tokens_a
                + [word_dict["[SEP]"]]
                + tokens_b
                + [word_dict["[SEP]"]]
            )

            segment_ids = [0] * (1 + len(tokens_a) + 1) + [1] * (len(tokens_b) + 1)

            # MASK LM
            n_pred = min(
                max_pred, max(1, int(round(len(input_ids) * 0.15)))
            )  # 15 % of tokens in one sentence(# 15% 的token需要进行预测)

            cand_maked_pos = [
                i
                for i, token in enumerate(input_ids)
                if token != word_dict["[CLS]"] and token != word_dict["[SEP]"]
            ]
            shuffle(cand_maked_pos)
            masked_tokens, masked_pos = [], []
            for pos in cand_maked_pos[:n_pred]:
                masked_pos.append(pos)
                masked_tokens.append(input_ids[pos])
                if random() < 0.8:  # 80%
                    input_ids[pos] = word_dict["[MASK]"]  # make mask
                elif random() < 0.5:  # 10%
                    index = randint(0, vocab_size - 1)  # random index in vocabulary
                    input_ids[pos] = word_dict[number_dict[index]]  # replace

            # Zero Paddings
            n_pad = maxlen - len(input_ids)
            input_ids.extend([0] * n_pad)
            segment_ids.extend([0] * n_pad)

            #     # Zero Padding (100% - 15%) tokens
            if max_pred > n_pred:
                n_pad = max_pred - n_pred
                masked_tokens.extend([0] * n_pad)
                masked_pos.extend([0] * n_pad)

            if tokens_a_index + 1 == tokens_b_index and positive < batch_size / 2:
                batch.append(
                    [input_ids, segment_ids, masked_tokens, masked_pos, True]
                )  # IsNext
                positive += 1
            elif tokens_a_index + 1 != tokens_b_index and negative < batch_size / 2:
                batch.append(
                    [input_ids, segment_ids, masked_tokens, masked_pos, False]
                )  # NotNext
                negative += 1
        return batch

    batch = make_batch()
    input_ids, segment_ids, masked_tokens, masked_pos, isNext = map(
        torch.LongTensor, zip(*batch, strict=False)
    )
    get_attn_pad_mask(input_ids, input_ids)[0][0], input_ids[0]

    # 多头注意力机制
    class MultiHeadAttention(nn.Module):
        def __init__(self):
            super(MultiHeadAttention, self).__init__()
            self.W_Q = nn.Linear(d_model, d_k * n_heads)
            self.W_K = nn.Linear(d_model, d_k * n_heads)
            self.W_V = nn.Linear(d_model, d_v * n_heads)

        def forward(self, Q, K, V, attn_mask):
            # q: [batch_size x len_q x d_model], k: [batch_size x len_k x d_model], v: [batch_size x len_k x d_model]
            residual, batch_size = Q, Q.size(0)
            # (B, S, D) -proj-> (B, S, D) -split-> (B, S, H, W) -trans-> (B, H, S, W)
            q_s = (
                self.W_Q(Q).view(batch_size, -1, n_heads, d_k).transpose(1, 2)
            )  # q_s: [batch_size x n_heads x len_q x d_k]
            k_s = (
                self.W_K(K).view(batch_size, -1, n_heads, d_k).transpose(1, 2)
            )  # k_s: [batch_size x n_heads x len_k x d_k]
            v_s = (
                self.W_V(V).view(batch_size, -1, n_heads, d_v).transpose(1, 2)
            )  # v_s: [batch_size x n_heads x len_k x d_v]

            attn_mask = attn_mask.unsqueeze(1).repeat(
                1, n_heads, 1, 1
            )  # attn_mask : [batch_size x n_heads x len_q x len_k]

            # context: [batch_size x n_heads x len_q x d_v], attn: [batch_size x n_heads x len_q(=len_k) x len_k(=len_q)]
            context, attn = ScaledDotProductAttention()(q_s, k_s, v_s, attn_mask)
            context = (
                context.transpose(1, 2).contiguous().view(batch_size, -1, n_heads * d_v)
            )  # context: [batch_size x len_q x n_heads * d_v]
            output = nn.Linear(n_heads * d_v, d_model)(context)
            return nn.LayerNorm(d_model)(
                output + residual
            ), attn  # output: [batch_size x len_q x d_model]

    # 嵌入层
    class Embedding(nn.Module):
        def __init__(self):
            super(Embedding, self).__init__()
            self.tok_embed = nn.Embedding(vocab_size, d_model)  # token 嵌入
            self.pos_embed = nn.Embedding(maxlen, d_model)  # 位置嵌入
            self.seg_embed = nn.Embedding(
                n_segments, d_model
            )  # segment(token type) embedding
            self.norm = nn.LayerNorm(d_model)

        def forward(self, x, seg):
            seq_len = x.size(1)
            pos = torch.arange(seq_len, dtype=torch.long)
            pos = pos.unsqueeze(0).expand_as(x)  # (seq_len,) -> (batch_size, seq_len)
            embedding = self.tok_embed(x) + self.pos_embed(pos) + self.seg_embed(seg)
            return self.norm(embedding)

    emb = Embedding()
    embeds = emb(input_ids, segment_ids)

    attenM = get_attn_pad_mask(input_ids, input_ids)

    class ScaledDotProductAttention(nn.Module):
        def __init__(self):
            super(ScaledDotProductAttention, self).__init__()

        def forward(self, Q, K, V, attn_mask):
            scores = torch.matmul(Q, K.transpose(-1, -2)) / np.sqrt(
                d_k
            )  # scores : [batch_size x n_heads x len_q(=len_k) x len_k(=len_q)]
            scores.masked_fill_(
                attn_mask, -1e9
            )  # Fills elements of self tensor with value where mask is one.
            attn = nn.Softmax(dim=-1)(scores)
            context = torch.matmul(attn, V)
            return context, attn

    SDPA = ScaledDotProductAttention()(embeds, embeds, embeds, attenM)

    # S, C, A = SDPA
    S, A = SDPA

    # print("Masks", masks[0][0])
    print()
    print("Scores: ", S[0][0], "\n\nAttention M: ", A[0][0])

    emb = Embedding()
    embeds = emb(input_ids, segment_ids)

    attenM = get_attn_pad_mask(input_ids, input_ids)

    MHA = MultiHeadAttention()(embeds, embeds, embeds, attenM)

    Output, A = MHA

    A[0][0]

    class PoswiseFeedForwardNet(nn.Module):
        def __init__(self):
            super(PoswiseFeedForwardNet, self).__init__()
            self.fc1 = nn.Linear(d_model, d_ff)
            self.fc2 = nn.Linear(d_ff, d_model)

        def forward(self, x):
            # (batch_size, len_seq, d_model) -> (batch_size, len_seq, d_ff) -> (batch_size, len_seq, d_model)
            return self.fc2(gelu(self.fc1(x)))

    # 编码器
    class EncoderLayer(nn.Module):
        def __init__(self):
            super(EncoderLayer, self).__init__()
            self.enc_self_attn = MultiHeadAttention()
            self.pos_ffn = PoswiseFeedForwardNet()

        def forward(self, enc_inputs, enc_self_attn_mask):
            enc_outputs, attn = self.enc_self_attn(
                enc_inputs, enc_inputs, enc_inputs, enc_self_attn_mask
            )  # enc_inputs to same Q,K,V
            enc_outputs = self.pos_ffn(
                enc_outputs
            )  # enc_outputs: [batch_size x len_q x d_model]
            return enc_outputs, attn

    class BERT(nn.Module):
        def __init__(self):
            super(BERT, self).__init__()
            self.embedding = Embedding()
            self.layers = nn.ModuleList([EncoderLayer() for _ in range(n_layers)])
            self.fc = nn.Linear(d_model, d_model)
            self.activ1 = nn.Tanh()
            self.linear = nn.Linear(d_model, d_model)
            self.activ2 = gelu
            self.norm = nn.LayerNorm(d_model)
            self.classifier = nn.Linear(d_model, 2)
            # decoder is shared with embedding layer
            embed_weight = self.embedding.tok_embed.weight
            n_vocab, n_dim = embed_weight.size()
            self.decoder = nn.Linear(n_dim, n_vocab, bias=False)
            self.decoder.weight = embed_weight
            self.decoder_bias = nn.Parameter(torch.zeros(n_vocab))

        def forward(self, input_ids, segment_ids, masked_pos):
            output = self.embedding(input_ids, segment_ids)
            enc_self_attn_mask = get_attn_pad_mask(input_ids, input_ids)
            for layer in self.layers:
                output, enc_self_attn = layer(output, enc_self_attn_mask)
            # output : [batch_size, len, d_model], attn : [batch_size, n_heads, d_mode, d_model]
            # it will be decided by first token(CLS)
            h_pooled = self.activ1(self.fc(output[:, 0]))  # [batch_size, d_model]
            logits_clsf = self.classifier(h_pooled)  # [batch_size, 2]

            masked_pos = masked_pos[:, :, None].expand(
                -1, -1, output.size(-1)
            )  # [batch_size, max_pred, d_model]
            # get masked position from final output of transformer.
            h_masked = torch.gather(
                output, 1, masked_pos
            )  # masking position [batch_size, max_pred, d_model]
            h_masked = self.norm(self.activ2(self.linear(h_masked)))
            logits_lm = (
                self.decoder(h_masked) + self.decoder_bias
            )  # [batch_size, max_pred, n_vocab]

            return logits_lm, logits_clsf

    logger.info("开始训练 bert 模型")
    model = BERT()
    criterion = nn.CrossEntropyLoss()
    optimizer = optim.Adam(model.parameters(), lr=0.001)

    batch = make_batch()
    input_ids, segment_ids, masked_tokens, masked_pos, isNext = map(
        torch.LongTensor, zip(*batch, strict=False)
    )

    for epoch in range(10):
        optimizer.zero_grad()
        logits_lm, logits_clsf = model(input_ids, segment_ids, masked_pos)
        loss_lm = criterion(logits_lm.transpose(1, 2), masked_tokens)  # for masked LM
        loss_lm = (loss_lm.float()).mean()
        loss_clsf = criterion(logits_clsf, isNext)  # for sentence classification
        loss = loss_lm + loss_clsf
        if (epoch + 1) % 10 == 0:
            print("Epoch:", "%04d" % (epoch + 1), "cost =", f"{loss:.6f}")
        loss.backward()
        optimizer.step()

    logger.info("开始使用模型")
    # Predict mask tokens ans isNext
    input_ids, segment_ids, masked_tokens, masked_pos, isNext = map(
        torch.LongTensor, zip(batch[0], strict=False)
    )
    logging.info("训练数据:")
    print(text)
    logging.info("处理的token 列表:")
    print(
        [
            number_dict[w.item()]
            for w in input_ids[0]
            if number_dict[w.item()] != "[PAD]"
        ]
    )

    logits_lm, logits_clsf = model(input_ids, segment_ids, masked_pos)
    logits_lm = logits_lm.data.max(2)[1][0].data.numpy()

    logger.info(
        "这表示在输入序列中被遮蔽（mask）的位置。这些位置的 token 将被模型用来预测 例如 15, 8, 26 是被遮蔽的 token 在输入序列中的索引位置。"
    )
    print(
        "masked tokens list : ",
        [pos.item() for pos in masked_tokens[0] if pos.item() != 0],
    )
    print(
        "对掩码部分预测的结果 它为空（[]），这可能表示模型在当前状态下没有成功预测遮蔽的 \n token predict masked tokens list : ",
        [pos for pos in logits_lm if pos != 0],
    )

    logits_clsf = logits_clsf.data.max(1)[1].data.numpy()[0]
    # isNext 表示实际标签（即真实数据中句子是否相邻）。
    print("isNext : ", True if isNext else False)
    # predict isNext 表示模型的预测结果。
    print("predict isNext : ", True if logits_clsf else False)



import logging

from fastapi import APIRouter

from mtmai.deps import SiteDep
from mtmai.models.webpage import PageMetaAuthor, PageMetaResponse

router = APIRouter()

logger = logging.getLogger()


@router.get("/page_meta", response_model=PageMetaResponse)
async def page_meta(
    *,
    site: SiteDep,
):
    logger.info("current site: %s", site)
    return PageMetaResponse(
        title="test_page_meta",
        description="test_page_meta",
        keywords=["test_page_meta"],
        authors=[PageMetaAuthor(name="test_page_meta", url="test_page_meta")],
        creator="test_page_meta",
        manifest="test_page_meta",
    )



"""
# 模型训练, 让llama3.18b模型具备将现代汉语转换为古文风格的文章
- 资料库: https://github.com/NiuTrans/Classical-Modern
- 参考别人的训练数据集： https://huggingface.co/datasets/AISPIN/shiji-70liezhuan
- 别人训练好的模型 参考： https://huggingface.co/AISPIN/Llama-3.1-8B-bnb-4bit-wenyanwen
- 参考视频： https://www.youtube.com/watch?v=Tq6qPw8EUVg

"""

import logging
import os

from fastapi import APIRouter

# from mtmlib.queue.queue import Message

# from mtmai.core.queue import Message

router = APIRouter()
logger = logging.getLogger()


def tran_guwen_consumer(msg: Message):
    import pandas as pd

    logger.info("start_tran_guwen")
    folder_path = r"双语数据/史记/七十列传"

    # get all subfolders in the folder, then for each subfolder, get source file and target file,
    # then read the content of the files then combine them into a dateset, the source file is the input field,
    # the target file is the output field, and the target file is the output field

    def get_files(folder_path):
        subfolders = [f.path for f in os.scandir(folder_path) if f.is_dir()]
        print(subfolders)
        dataset = []

        source_file = "source.txt"
        target_file = "target.txt"
        for x in subfolders:
            with open(os.path.join(x, source_file), encoding="utf-8") as f:
                source_content = f.read()
            with open(os.path.join(x, target_file), encoding="utf-8") as f:
                target_content = f.read()

            # source and target needs to be split by "\n"
            source_content = source_content.split("\n")
            target_content = target_content.split("\n")

            # source and target should be saved into dateset line by line
            for i in range(len(source_content)):
                dataset.append([source_content[i], target_content[i]])

        return dataset

    dataset = get_files(folder_path)

    # add one column "instruction" with the content "请把古文翻译成现代汉语" to the dataset
    df = pd.DataFrame(dataset, columns=["source", "target"])
    df["instruction"] = "请把现代汉语翻译成古文"

    # rename the columns: source -> output, target -> input
    df.rename(columns={"source": "output", "target": "input"}, inplace=True)

    # print length of the dataset
    print(len(df))

    # save the dataset into a jsonl file
    df.to_json("dataset.jsonl", orient="records", lines=True, force_ascii=False)



import json
import shutil
from pathlib import Path

from langchain.tools import tool


class TemplateTools:  # noqa: D101
    @tool("Learn landing page options")
    def learn_landing_page_options(input):
        """Learn the templates at your disposal."""
        templates = json.load(open("config/templates.json"))
        return json.dumps(templates, indent=2)

    @tool("Copy landing page template to project folder")
    def copy_landing_page_template_to_project_folder(landing_page_template):
        """Copy a landing page template to your project
        folder so you can start modifying it, it expects
        a landing page template folder as input.
        """  # noqa: D205
        source_path = Path(f"templates/{landing_page_template}")
        destination_path = Path(f"workdir/{landing_page_template}")
        destination_path.parent.mkdir(parents=True, exist_ok=True)
        shutil.copytree(source_path, destination_path)
        return f"Template copied to {landing_page_template} and ready to be modified, main files should be under ./{landing_page_template}/src/components, you should focus on those."



import logging

from langgraph.graph import END, StateGraph

from mtmai.agents.conditions import condition_error_edge
from mtmai.agents.nodes.conduct_interviews_node import ConductInterviewNode
from mtmai.agents.nodes.index_references_node import IndexReferencesNode
from mtmai.agents.nodes.initialize_research_node import InitializeResearchNode
from mtmai.agents.nodes.refine_outline_node import RefineOutlineNode
from mtmai.agents.nodes.write_article_node import WriteArticleNode
from mtmai.agents.nodes.write_sections_node import WriteSectionsNode
# from mtmai.llm.llm import get_llm_chatbot_default
from mtmai.models.graph_config import ResearchState

logger = logging.getLogger()


class StormGraph():
    def create_graph(self) -> StateGraph:
        # wf = StateGraph(MainState)
        # wf.add_node("human_input", HumanInputNode())
        # child_graph = self.create_sub_graph_storm().compile()

        # async def call_storm_graph(
        #     state: MainState, config: RunnableConfig
        # ) -> MainState:
        #     try:
        #         child_graph_input = {"topic": state.get("prompt")}
        #         child_graph_output = await child_graph.ainvoke(
        #             child_graph_input, config
        #         )

        #         # 输出的结果是文章生成结果
        #         logger.info(f"文章生成 结果: {child_graph_output}")

        #         return {
        #             "messages": [
        #                 ChatMessage(role="assistant", content=state.get("生成结束"))
        #             ]
        #         }
        #     except Exception as e:
        #         logger.error(f"生成文章失败: {e}")
        #         import traceback

        #         error_message = (
        #             f"生成文章失败: {e}\n\nDebug Stack:\n{traceback.format_exc()}"
        #         )
        #         logger.error(error_message)
        #         return {
        #             "messages": [ChatMessage(role="assistant", content="生成文章失败")]
        #         }

        # wf.add_node("storm", call_storm_graph)
        # wf.set_entry_point("human_input")
        # wf.add_edge("human_input", "storm")
        # wf.add_edge("storm", END)

        # return wf
        return self.create_sub_graph_storm()

    def create_sub_graph_storm(self):
        """创建 storm 子图"""
        wf = StateGraph(ResearchState)
        # llm = get_llm_chatbot_default()
        wf.add_node("init_research", InitializeResearchNode())
        wf.add_node("conduct_interviews", ConductInterviewNode())
        wf.add_node("refine_outline", RefineOutlineNode())
        wf.add_node("index_references", IndexReferencesNode())
        wf.add_node("write_sections", WriteSectionsNode())
        wf.add_node("write_article", WriteArticleNode())
        # 添加边
        wf.add_conditional_edges(
            "init_research",
            condition_error_edge,
            {
                "continue": "conduct_interviews",
                "error": END,
            },
        )
        wf.add_conditional_edges(
            "conduct_interviews",
            condition_error_edge,
            {
                "continue": "refine_outline",
                "error": END,
            },
        )
        wf.add_conditional_edges(
            "refine_outline",
            condition_error_edge,
            {
                "continue": "index_references",
                "error": END,
            },
        )
        wf.add_edge("index_references", "write_sections")
        wf.add_edge("write_sections", "write_article")
        wf.add_edge("write_article", END)

        wf.set_entry_point("init_research")

        return wf



import logging
import os
import time

logger = logging.getLogger()

# from langchain_community.agent_toolkits import GmailToolkit
# from langchain_community.tools.gmail.search import GmailSearch


class Nodes:
    def __init__(self):
        # self.gmail = GmailToolkit()
        pass

    def check_email(self, state):
        logger.info("# Checking for new emails")
        # search = GmailSearch(api_resource=self.gmail.api_resource)
        # emails = search("after:newer_than:1d")
        emails = []
        checked_emails = (
            state["checked_emails_ids"] if state["checked_emails_ids"] else []
        )
        thread = []
        new_emails = []
        for email in emails:
            if (
                (email["id"] not in checked_emails)
                and (email["threadId"] not in thread)
                and (os.environ["MY_EMAIL"] not in email["sender"])
            ):
                thread.append(email["threadId"])
                new_emails.append(
                    {
                        "id": email["id"],
                        "threadId": email["threadId"],
                        "snippet": email["snippet"],
                        "sender": email["sender"],
                    }
                )

        ids_to_check = ["id1", "id2"]
        # checked_emails.extend([email["id"] for email in emails])
        checked_emails.extend(ids_to_check)
        return {**state, "emails": new_emails, "checked_emails_ids": checked_emails}

    def wait_next_run(self, state):
        logger.debug("## wait_next_run")
        time.sleep(2)
        return state

    def new_emails(self, state):
        if len(state["emails"]) == 0:
            logger.info("## No new emails")
            return "end"
        else:
            print("## New emails")
            return "continue"



from crewai import Crew
from langchain_core.runnables import Runnable

from mtmai.llm.llm import get_llm_chatbot_default

from .agents import JokAgents
from .tasks import MyTasks


class JokeCrew:
    def __init__(self, runnable: Runnable):
        agents = JokAgents()
        # self.filter_agent = agents.email_filter_agent()
        self.joke_writer_agent = agents.joke_writer_agent()
        # self.writer_agent = agents.email_response_writer()
        self.runnable = runnable

    async def kickoff(self, state):
        tasks = MyTasks()
        # emails = ["a@a.com"]
        llm = get_llm_chatbot_default()
        crew = Crew(
            agents=[
                self.joke_writer_agent,
                # self.action_agent,
                # self.writer_agent,
            ],
            tasks=[
                tasks.task_demo_human(self.joke_writer_agent),
                # tasks.action_required_emails_task(self.action_agent),
                # tasks.draft_responses_task(self.writer_agent),
            ],
            verbose=True,
            manager_llm=llm,
            function_calling_llm=llm,
            planning_llm=llm,
        )
        result = await crew.kickoff_async()

        artifact = (
            {
                "artiface_type": "Document",
                "title": "任务结果",
                "props": {"title": "任务结果", "content": result},
            },
        )
        return {**state, "artifacts": [artifact]}



from typing import cast

from mtmai.agents.ctx import init_mtmai_context
from mtmai.worker import wfapp
from mtmai.workflows.flow_crewai import FlowCrewAIAgent
from mtmaisdk.clients.rest.models import AgentNodeRunRequest, CrewAIParams
from mtmaisdk.context.context import Context


@wfapp.workflow(
    on_events=["agent:call"],
    input_validator=AgentNodeRunRequest,
)
class FlowRecovery:
    """FlowRecovery"""

    @wfapp.step(timeout="10m", retries=1)
    async def run(self, hatctx: Context):
        init_mtmai_context(hatctx)
        input = cast(AgentNodeRunRequest, hatctx.workflow_input())
        hatctx.log(f"input: {input}")

        if input.flow_name == "crewai":
            params = CrewAIParams()
            hatctx.log("调用子工作流")
            try:
                await hatctx.aio.spawn_workflow(
                    FlowCrewAIAgent.__name__, params.model_dump()
                )
            except Exception as e:
                # Spawn a recovery workflow
                await hatctx.aio.spawn_workflow("recovery-workflow", {"error": str(e)})
            return {"next": "step_a1"}
        elif input.flow_name == "scrape":
            return {"next": "step_b1"}
        # return await StepCallAgent(hatctx).run()
        return {
            "next": "step_a1",
        }



import hashlib
import itertools
import os
import platform
import random
import time

# 6/20/2022 12AM
BASE_EPOCH = 1655683200
VERSION = 0

# Number of bits
TIMESTAMP_BITS = 32
WORKER_ID_BITS = 21
SEQUENCE_BITS = 10
VERSION_BITS = 1

# Bit shits (left)
TIMESTAMP_SHIFT = 32
WORKER_ID_SHIFT = 11
SEQUENCE_SHIFT = 1
VERSION_SHIFT = 0

SEQUENCE_MAX = (2**SEQUENCE_BITS) - 1
_sequence_start = None
SEQUENCE_COUNTER = itertools.count()
_worker_hash = None

# prefix
ORGANIZATION_AUTH_TOKEN_PREFIX = "oat"
ORG_PREFIX = "o"
TASK_PREFIX = "tsk"
USER_PREFIX = "u"
STEP_PREFIX = "stp"
ARTIFACT_PREFIX = "a"
WORKFLOW_PREFIX = "w"
WORKFLOW_PERMANENT_ID_PREFIX = "wpid"
WORKFLOW_RUN_PREFIX = "wr"
WORKFLOW_PARAMETER_PREFIX = "wp"
AWS_SECRET_PARAMETER_PREFIX = "asp"
OUTPUT_PARAMETER_PREFIX = "op"
BITWARDEN_LOGIN_CREDENTIAL_PARAMETER_PREFIX = "blc"
BITWARDEN_SENSITIVE_INFORMATION_PARAMETER_PREFIX = "bsi"
BITWARDEN_CREDIT_CARD_DATA_PARAMETER_PREFIX = "bccd"
TASK_GENERATION_PREFIX = "tg"


def generate_workflow_id() -> str:
    int_id = generate_id()
    return f"{WORKFLOW_PREFIX}_{int_id}"


def generate_workflow_permanent_id() -> str:
    int_id = generate_id()
    return f"{WORKFLOW_PERMANENT_ID_PREFIX}_{int_id}"


def generate_workflow_run_id() -> str:
    int_id = generate_id()
    return f"{WORKFLOW_RUN_PREFIX}_{int_id}"


def generate_aws_secret_parameter_id() -> str:
    int_id = generate_id()
    return f"{AWS_SECRET_PARAMETER_PREFIX}_{int_id}"


def generate_workflow_parameter_id() -> str:
    int_id = generate_id()
    return f"{WORKFLOW_PARAMETER_PREFIX}_{int_id}"


def generate_output_parameter_id() -> str:
    int_id = generate_id()
    return f"{OUTPUT_PARAMETER_PREFIX}_{int_id}"


def generate_bitwarden_login_credential_parameter_id() -> str:
    int_id = generate_id()
    return f"{BITWARDEN_LOGIN_CREDENTIAL_PARAMETER_PREFIX}_{int_id}"


def generate_bitwarden_sensitive_information_parameter_id() -> str:
    int_id = generate_id()
    return f"{BITWARDEN_SENSITIVE_INFORMATION_PARAMETER_PREFIX}_{int_id}"


def generate_bitwarden_credit_card_data_parameter_id() -> str:
    int_id = generate_id()
    return f"{BITWARDEN_CREDIT_CARD_DATA_PARAMETER_PREFIX}_{int_id}"


def generate_organization_auth_token_id() -> str:
    int_id = generate_id()
    return f"{ORGANIZATION_AUTH_TOKEN_PREFIX}_{int_id}"


def generate_org_id() -> str:
    int_id = generate_id()
    return f"{ORG_PREFIX}_{int_id}"


def generate_task_id() -> str:
    int_id = generate_id()
    return f"{TASK_PREFIX}_{int_id}"


def generate_step_id() -> str:
    int_id = generate_id()
    return f"{STEP_PREFIX}_{int_id}"


def generate_artifact_id() -> str:
    int_id = generate_id()
    return f"{ARTIFACT_PREFIX}_{int_id}"

def generate_mttask_id() -> str:
    int_id = generate_id()
    return f"{TASK_PREFIX}_{int_id}"


def generate_user_id() -> str:
    int_id = generate_id()
    return f"{USER_PREFIX}_{int_id}"


def generate_task_generation_id() -> str:
    int_id = generate_id()
    return f"{TASK_GENERATION_PREFIX}_{int_id}"


def generate_totp_code_id() -> str:
    int_id = generate_id()
    return f"totp_{int_id}"


def generate_action_id() -> str:
    int_id = generate_id()
    return f"a_{int_id}"


def generate_id() -> int:
    """
    generate a 64-bit int ID
    """
    create_at = current_time() - BASE_EPOCH
    sequence = _increment_and_get_sequence()

    time_part = _mask_shift(create_at, TIMESTAMP_BITS, TIMESTAMP_SHIFT)
    worker_part = _mask_shift(_get_worker_hash(), WORKER_ID_BITS, WORKER_ID_SHIFT)
    sequence_part = _mask_shift(sequence, SEQUENCE_BITS, SEQUENCE_SHIFT)
    version_part = _mask_shift(VERSION, VERSION_BITS, VERSION_SHIFT)

    return time_part | worker_part | sequence_part | version_part


def _increment_and_get_sequence() -> int:
    global _sequence_start
    if _sequence_start is None:
        _sequence_start = random.randint(0, SEQUENCE_MAX)

    return (_sequence_start + next(SEQUENCE_COUNTER)) % SEQUENCE_MAX


def current_time() -> int:
    return int(time.time())


def current_time_ms() -> int:
    return int(time.time() * 1000)


def _mask_shift(value: int, mask_bits: int, shift_bits: int) -> int:
    return (value & ((2**mask_bits) - 1)) << shift_bits


def _get_worker_hash() -> int:
    global _worker_hash
    if _worker_hash is None:
        _worker_hash = _generate_worker_hash()
    return _worker_hash


def _generate_worker_hash() -> int:
    worker_identity = f"{platform.node()}:{os.getpid()}"
    return int(hashlib.md5(worker_identity.encode()).hexdigest()[-15:], 16)



# coding: utf-8

"""
    Mtmai API

    The Mtmai API

    The version of the OpenAPI document: 1.0.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501

import warnings
from pydantic import validate_call, Field, StrictFloat, StrictStr, StrictInt
from typing import Any, Dict, List, Optional, Tuple, Union
from typing_extensions import Annotated

from pydantic import Field
from typing_extensions import Annotated
from mtmai.gomtmclients.rest.models.list_slack_webhooks import ListSlackWebhooks

from mtmai.gomtmclients.rest.api_client import ApiClient, RequestSerialized
from mtmai.gomtmclients.rest.api_response import ApiResponse
from mtmai.gomtmclients.rest.rest import RESTResponseType


class SlackApi:
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None) -> None:
        if api_client is None:
            api_client = ApiClient.get_default()
        self.api_client = api_client


    @validate_call
    async def slack_webhook_delete(
        self,
        slack: Annotated[str, Field(min_length=36, strict=True, max_length=36, description="The Slack webhook id")],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> None:
        """Delete Slack webhook

        Delete Slack webhook

        :param slack: The Slack webhook id (required)
        :type slack: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._slack_webhook_delete_serialize(
            slack=slack,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '204': None,
            '400': "APIErrors",
            '401': "APIErrors",
            '405': "APIErrors",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        await response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data


    @validate_call
    async def slack_webhook_delete_with_http_info(
        self,
        slack: Annotated[str, Field(min_length=36, strict=True, max_length=36, description="The Slack webhook id")],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[None]:
        """Delete Slack webhook

        Delete Slack webhook

        :param slack: The Slack webhook id (required)
        :type slack: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._slack_webhook_delete_serialize(
            slack=slack,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '204': None,
            '400': "APIErrors",
            '401': "APIErrors",
            '405': "APIErrors",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        await response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )


    @validate_call
    async def slack_webhook_delete_without_preload_content(
        self,
        slack: Annotated[str, Field(min_length=36, strict=True, max_length=36, description="The Slack webhook id")],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Delete Slack webhook

        Delete Slack webhook

        :param slack: The Slack webhook id (required)
        :type slack: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._slack_webhook_delete_serialize(
            slack=slack,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '204': None,
            '400': "APIErrors",
            '401': "APIErrors",
            '405': "APIErrors",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        return response_data.response


    def _slack_webhook_delete_serialize(
        self,
        slack,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:

        _host = None

        _collection_formats: Dict[str, str] = {
        }

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[
            str, Union[str, bytes, List[str], List[bytes], List[Tuple[str, bytes]]]
        ] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        if slack is not None:
            _path_params['slack'] = slack
        # process the query parameters
        # process the header parameters
        # process the form parameters
        # process the body parameter


        # set the HTTP header `Accept`
        if 'Accept' not in _header_params:
            _header_params['Accept'] = self.api_client.select_header_accept(
                [
                    'application/json'
                ]
            )


        # authentication setting
        _auth_settings: List[str] = [
            'basicAuth', 
            'cookieAuth', 
            'bearerAuth'
        ]

        return self.api_client.param_serialize(
            method='DELETE',
            resource_path='/api/v1/slack/{slack}',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth
        )




    @validate_call
    async def slack_webhook_list(
        self,
        tenant: Annotated[str, Field(min_length=36, strict=True, max_length=36, description="The tenant id")],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ListSlackWebhooks:
        """List Slack integrations

        List Slack webhooks

        :param tenant: The tenant id (required)
        :type tenant: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._slack_webhook_list_serialize(
            tenant=tenant,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "ListSlackWebhooks",
            '400': "APIErrors",
            '401': "APIErrors",
            '405': "APIErrors",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        await response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data


    @validate_call
    async def slack_webhook_list_with_http_info(
        self,
        tenant: Annotated[str, Field(min_length=36, strict=True, max_length=36, description="The tenant id")],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[ListSlackWebhooks]:
        """List Slack integrations

        List Slack webhooks

        :param tenant: The tenant id (required)
        :type tenant: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._slack_webhook_list_serialize(
            tenant=tenant,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "ListSlackWebhooks",
            '400': "APIErrors",
            '401': "APIErrors",
            '405': "APIErrors",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        await response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )


    @validate_call
    async def slack_webhook_list_without_preload_content(
        self,
        tenant: Annotated[str, Field(min_length=36, strict=True, max_length=36, description="The tenant id")],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """List Slack integrations

        List Slack webhooks

        :param tenant: The tenant id (required)
        :type tenant: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._slack_webhook_list_serialize(
            tenant=tenant,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "ListSlackWebhooks",
            '400': "APIErrors",
            '401': "APIErrors",
            '405': "APIErrors",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        return response_data.response


    def _slack_webhook_list_serialize(
        self,
        tenant,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:

        _host = None

        _collection_formats: Dict[str, str] = {
        }

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[
            str, Union[str, bytes, List[str], List[bytes], List[Tuple[str, bytes]]]
        ] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        if tenant is not None:
            _path_params['tenant'] = tenant
        # process the query parameters
        # process the header parameters
        # process the form parameters
        # process the body parameter


        # set the HTTP header `Accept`
        if 'Accept' not in _header_params:
            _header_params['Accept'] = self.api_client.select_header_accept(
                [
                    'application/json'
                ]
            )


        # authentication setting
        _auth_settings: List[str] = [
            'basicAuth', 
            'cookieAuth', 
            'bearerAuth'
        ]

        return self.api_client.param_serialize(
            method='GET',
            resource_path='/api/v1/tenants/{tenant}/slack',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth
        )





import json
import re
import uuid

import orjson
from json_repair import repair_json
from langchain_core.messages import AIMessage, ToolCall
from langgraph.checkpoint.memory import MemorySaver

# from mtmai.llm.llm import get_llm_chatbot_default

memory = MemorySaver()


async def call_model(
    prompt: list,
    model: str,
    max_retries: int = 2,
    response_format: str | None = None,
    api_key: str | None = None,
) -> str:
    """
    这个函数原本 仅能使用 chatgpt 模型，并且要求返回格式必须是 json
    现在暂时用 from together import Together 替代。
    """
    optional_params = {}
    if response_format == "json":
        optional_params = {"response_format": {"type": "json_object"}}

    # search = TavilySearchResults(max_results=2)
    tools = []
    # agent_executor = create_react_agent(
    #     model,
    #     # tools,
    #     checkpointer=memory,
    # )
    # llm = get_llm_chatbot_default()
    # response = await llm.ainvoke(prompt)
    # print(response)
    response.content
    return response.content

    # llmget_llm_chatbot_default()
    # lc_messages = convert_openai_messages(prompt)

    # llm = lcllm_openai_chat()
    # response = (
    #     # ChatOpenAI(
    #     #     model=model,
    #     #     max_retries=max_retries,
    #     #     model_kwargs=optional_params,
    #     #     api_key=api_key,
    #     # )
    #     llm.invoke(lc_messages).content
    # )
    # together = Together()
    # extract = together.chat.completions.create(
    #     messages=prompt,
    #     model="meta-llama/Meta-Llama-3.1-8B-Instruct-Turbo",
    #     # response_format=optional_params,
    #     response_format={
    #         "type": "json_object",
    #         # "schema": VoiceNote.model_json_schema(),
    #     },
    # )

    # extract_json = extract_json_from_string(extract.choices[0].message.content)
    # return extract_json
    # output = json.loads(extract.choices[0].message.content)
    # print(json.dumps(output, indent=2))
    # return output


def extract_json_from_string(json_str: str) -> dict:
    # 清理 JSON 字符串
    cleaned_json_str = clean_json_string(json_str)
    if cleaned_json_str.startswith("{"):
        return cleaned_json_str
    match = re.search(r"```.*?({[\s\S]*?})\s*```", cleaned_json_str)
    if match:
        json_str = match.group(1)
        return json_str
    print(f"未找到 JSON 数据 : {cleaned_json_str}")
    return None


def clean_json_string(json_str: str) -> str:
    # 替换非法控制字符
    json_str = re.sub(r"[\x00-\x1f\x7f]", "", json_str)
    return json_str


def fix_tool_calls(ai_msg: AIMessage):
    if not ai_msg.content or (
        ai_msg.tool_calls and ai_msg.tool_calls[0].type == "tool_call"
    ):
        # 已是正确格式
        return ai_msg
    # 情况1： 以lamma3.1 常见的 <function> 格式回复函数函数调用
    function_regex = r"<function=(\w+)>(.*?)</function>"
    match = re.search(function_regex, ai_msg.content, re.DOTALL)
    if match:
        function_name, args_string = match.groups()

        args = json.loads(args_string)
        ai_msg.tool_calls.append(
            ToolCall(
                name=function_name,
                arguments=args,
                id=str(uuid.uuid4()),
                type="tool_call",
            )
        )
        ai_msg.content = ""
        return ai_msg

    # 情况2 函数调用没有出现在 tool_calls 字段，而是出现在 content 中，json格式
    # 例子： {"name": "ToDevelopAssistant", "parameters": {"request": "I want to edit an article"}}
    loaded_data = orjson.loads(repair_json(ai_msg.content))
    if (
        isinstance(loaded_data, dict)
        and "name" in loaded_data
        and "parameters" in loaded_data
    ):
        ai_msg.tool_calls.append(
            ToolCall(
                name=loaded_data["name"],
                arguments=loaded_data["parameters"],
                id=str(uuid.uuid4()),
                type="tool_call",
            )
        )
        ai_msg.content = ""
        return ai_msg
    return ai_msg



"""HTML processing functions"""
from __future__ import annotations

from bs4 import BeautifulSoup
from requests.compat import urljoin


def extract_hyperlinks(soup: BeautifulSoup, base_url: str) -> list[tuple[str, str]]:
    """Extract hyperlinks from a BeautifulSoup object

    Args:
        soup (BeautifulSoup): The BeautifulSoup object
        base_url (str): The base URL

    Returns:
        List[Tuple[str, str]]: The extracted hyperlinks
    """
    return [
        (link.text, urljoin(base_url, link["href"]))
        for link in soup.find_all("a", href=True)
    ]


def format_hyperlinks(hyperlinks: list[tuple[str, str]]) -> list[str]:
    """Format hyperlinks to be displayed to the user

    Args:
        hyperlinks (List[Tuple[str, str]]): The hyperlinks to format

    Returns:
        List[str]: The formatted hyperlinks
    """
    return [f"{link_text} ({link_url})" for link_text, link_url in hyperlinks]



"""
doc: https://sdk.vercel.ai/docs/ai-sdk-ui/stream-protocol
"""

import json

from fastapi.encoders import jsonable_encoder
from fastapi.responses import StreamingResponse


def text(word: str):
    return f"0:{json.dumps(word)}\n"


def data(items):
    if isinstance(items, list):
        return f"2:{json.dumps(jsonable_encoder(items))}\n"
    else:
        return f"2:{json.dumps([jsonable_encoder(items)])}\n"


def error(error_message: str):
    return f"3:{json.dumps(error_message)}\n"


def finish(reason: str = "stop", prompt_tokens: int = 0, completion_tokens: int = 0):
    data = {
        "finishReason": reason,
        "usage": {"promptTokens": prompt_tokens, "completionTokens": completion_tokens},
    }
    return f"d:{json.dumps(data)}\n"


def AiSDKStreamResponse(content):
    """
    说明: 在 cloudflared tunnel 上，是支持 http sse 的，不需要特别添加 headers。
    例子：
    @router.get("/hello_stream1")
    async def hello_stream1():
        from fastapi.responses import StreamingResponse
        async def event_generator():
            for i in range(10):
                yield aisdk.text(f"Hello, this is message {i}")
                await asyncio.sleep(0.5)

        return StreamingResponse(event_generator(), media_type="text/event-stream")

    **注意** 如果加了多余的http header 可能导致 cloudflared tunnel 环境下无法正常工作。
    """
    return StreamingResponse(
        content,
        media_type="text/event-stream",
        # headers={
        #     "Cache-Control": "no-cache, no-transform, no-store, must-revalidate, proxy-revalidate",
        #     "x-vercel-ai-data-stream": "v1",
        #     "content-type": "text/plain; charset=utf-8",
        #     "vary": "RSC, Next-Router-State-Tree, Next-Router-Prefetch",
        #     "cache-control": "no-store, no-cache, must-revalidate, proxy-revalidate",
        #     "pragma": "no-cache",
        #     "expires": "0",
        #     "surrogate-control": "no-store",
        #     "Content-Type": "text/event-stream",
        #     "Connection": "keep-alive",
        #     "X-Accel-Buffering": "no",  # 用于 Nginx 代理
        #     "CF-Cache-Status": "DYNAMIC",  # 告诉 Cloudflare 这是动态内容
        #     "CF-Edge-Cache-Status": "DYNAMIC",  # 告诉 Cloudflare 这是动态内容
        #     "Transfer-Encoding": "chunked",  # 明确指定分块传输编码
        #     "X-Content-Type-Options": "nosniff",  # 防止内容类型嗅探
        # },
    )


# -------------------------------------------------------------------------
class AiTextChunck:
    def __init__(self, text: str):
        self.text = text

    def __str__(self):
        return self.text



import uuid
from datetime import datetime
from typing import Literal

from pydantic import BaseModel, HttpUrl
from pydantic import Field as PydanticField
from sqlmodel import JSON, Column, Field, SQLModel


class SitePublishConfig(BaseModel):
    """
    站点发布配置
    """

    publishType: Literal["local", "remove"] = Field(default="local")


class SiteHostBase(SQLModel):
    domain: str = Field(min_length=3, max_length=255)
    is_default: bool = Field(default=False)
    is_https: bool = Field(default=False)
    site_id: uuid.UUID = Field(
        foreign_key="site.id", nullable=False, ondelete="CASCADE"
    )


class SiteHost(SiteHostBase, table=True):
    id: uuid.UUID = Field(default_factory=uuid.uuid4, primary_key=True)


class SiteBase(SQLModel):
    """
    用户站点基础配置
    """

    title: str | None = Field(default=None, max_length=255)
    description: str | None = Field(default=None, max_length=255)
    created_at: datetime = Field(default=datetime.now())
    updated_at: datetime = Field(default=datetime.now())
    enabled: bool | None = Field(default=True)
    ########################################################################################
    # 新设计方式：使用第三方站点的方式，例如可以绑定 wordpress 站点
    url: str | None = Field(default=None, max_length=255)
    # 站点的框架，例如 wordpress, typecho, hexo 等
    framework: str | None = Field(default=None, max_length=255)
    # 站点的认证信息，例如 wordpress 的 username, password,
    credential_type: str | None = Field(default=None, max_length=255)
    credentials: str | None = Field(default=None, max_length=255)
    meta: dict | None = Field(default={}, sa_column=Column(JSON))
    enabled_automation: bool | None = Field(default=False)

    # siteGenConfigActiveId: uuid.UUID | None = Field(default=None)
    owner_id: uuid.UUID = Field(
        foreign_key="user.id", index=True, nullable=False, ondelete="CASCADE"
    )


# Database model, database table inferred from class name
class Site(SiteBase, table=True):
    id: uuid.UUID = Field(default_factory=uuid.uuid4, primary_key=True)
    updated_at: datetime = Field(default=datetime.now())
    # site_autos: List["TaskSchedule"] = Relationship(back_populates="site")


class SiteAutoConfig(BaseModel):
    agent_name: str = Field(default="auto_agent", description="agent 名称")
    limit: int = Field(default=10, description="每次生成文章的数量")
    prompt: str = Field(default="", description="生成文章的额外提示词")
    keywords: str | None = Field(default=None, max_length=255, description="关键词")
    sitePublishConfig: SitePublishConfig = Field(default={}, description="站点发布配置")


class SiteCreateRequest(BaseModel):
    url: HttpUrl = PydanticField(
        default=None,
        max_length=255,
        description="网站地址",
        # json_schema_extra={
        #     "label": "网站地址",
        #     "placeholder": "https://example.com",
        #     "icon": "site",
        # },
    )

    class Config:
        json_schema_extra = {
            "title": "站点创建",
        }


class SiteUpdateRequest(SiteBase):
    owner_id: uuid.UUID | None = None


class SiteItemPublic(SiteBase):
    id: uuid.UUID


class ListSiteResponse(SQLModel):
    data: list[SiteItemPublic]
    count: int


class ListSiteHostRequest(SQLModel):
    siteId: uuid.UUID
    q: str | None = Field(default=None, max_length=255)


class SiteHostItemPublic(SiteHostBase):
    id: uuid.UUID


class ListSiteHostsResponse(SQLModel):
    data: list[SiteHostItemPublic]
    count: int


class SiteHostCreateRequest(SiteHostBase):
    site_id: uuid.UUID


class SiteHostCreateResponse(SQLModel):
    id: uuid.UUID


class SiteHostUpdateRequest(SiteHostBase):
    id: uuid.UUID
    host: str = Field(min_length=3, max_length=255)


class SiteHostUpdateResponse(SQLModel):
    id: uuid.UUID


class SiteHostDeleteRequest(SQLModel):
    id: uuid.UUID


class SiteHostDeleteResponse(SQLModel):
    id: uuid.UUID



import logging

from autogen_agentchat.agents import AssistantAgent
from autogen_agentchat.conditions import MaxMessageTermination, TextMentionTermination
from autogen_agentchat.teams import RoundRobinGroupChat
from autogen_core.tools import FunctionTool

from ...tools.calculator import web_search

# from ..agents.ag.model_client import get_oai_Model
# from ..models.ag import TeamResult
# from ..tools.calculator import web_search
from .model_client import get_oai_Model

logger = logging.getLogger(__name__)


class TeamBuilder:
    """Manages team operations including loading configs and running teams"""

    async def create_demo_team(self):
        """创建默认测试团队"""
        base_model = get_oai_Model()
        calculator_fn_tool = FunctionTool(
            name="calculator",
            description="A simple calculator that performs basic arithmetic operations",
            func=web_search,
            global_imports=[],
        )

        calc_assistant = AssistantAgent(
            name="assistant_agent",
            system_message="You are a helpful assistant. Solve tasks carefully. When done, say TERMINATE.",
            model_client=base_model,
            tools=[calculator_fn_tool],
        )
        # Create termination conditions for calculator team
        calc_text_term = TextMentionTermination(text="TERMINATE")
        calc_max_term = MaxMessageTermination(max_messages=10)
        calc_or_term = calc_text_term | calc_max_term
        calc_or_term = calc_text_term | calc_max_term
        calc_team = RoundRobinGroupChat(
            participants=[calc_assistant], termination_condition=calc_or_term
        )
        return calc_team
